! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010-2012 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!        MPI datatype constructor for data field arrays        !
!                                                              !
!--------------------------------------------------------------!

#include "fd4.h"

!! MPI datatype constructor for data field arrays
!!
!! Preprocessor options:
!!  - $FD4_NO_MPI_TYPE_CREATE_SUBARRAY: Do not use MPI_Type_create_subarray (in case no MPI 2 present)
!!
!! @author Matthias Lieber
module fd4_comm_mod

  use fd4_globaldef_mod
  use fd4_mpi_mod

  USE_MPI
  
  implicit none
  private

  INCLUDE_MPIFH
  
  public :: fd4_comm_build_array_type
  
  integer(MPI_ADDRESS_KIND), allocatable :: displ(:)
  integer, allocatable :: blens(:)
  integer, allocatable :: types(:)
  
contains
  

  !! Build an MPI datatype for a single 4D array using a stride on the 1st and
  !! a subset of the other three dimensions.
  !!
  !! The stride on 1st dimension is defined with $stride, e.g.:
  !!> 
  !!    stride(1,1:2) = (/ 1,2 /)   ! start end end index of one contiguous subsection
  !!    stride(2,1:2) = (/ 4,4 /)
  !!    stride(3,1:2) = (/ 7,9 /)
  !!<
  !! selects the following values from dimension 0 of the array, assuming size 9:
  !!>
  !!    123456789
  !!    xx x  xxx
  !!<
  !!
  !! The implementation uses $MPI_Type_create_subarray (MPI-2) unless $FD4_NO_MPI_TYPE_CREATE_SUBARRAY is set.
  !! The resulting data type is not commited!
  !!
  subroutine fd4_comm_build_array_type(ext, stride, sbnd, sdata_type, err)
    integer(i_k), intent(in)      :: ext(0:3)       !! extents of the sdata array
    integer(i_k), intent(in)      :: stride(:,:)    !! stride to select subsets of the array for the datatype
    integer(i_k), intent(in)      :: sbnd(3,2)      !! bounds to select a spatial subset
    integer, intent(out)          :: sdata_type     !! the MPI datatype for the array
    integer(i_k), intent(out)     :: err            !! error status: 0...ok

    integer :: i, count
#ifndef FD4_NO_MPI_TYPE_CREATE_SUBARRAY
    integer, parameter :: ndims=4
    integer :: array_of_subsizes(0:ndims-1), array_of_starts(0:ndims-1)
#else
    integer :: j, k, cell_type
#endif

    err = -1

#ifndef FD4_NO_MPI_TYPE_CREATE_SUBARRAY
    
    !! using MPI_Type_create_subarray (prefered version)

    ! number of subsections in 1st dim
    count = ubound(stride,1)
    array_of_subsizes(1:3) = sbnd(1:3,2) - sbnd(1:3,1) + 1
    array_of_starts(1:3)   = sbnd(1:3,1) - 1
    
    call fd4_comm_alloc_helpers(count)

    ! build type for each subsection
    do i=1,count
      array_of_subsizes(0) = stride(i,2) - stride(i,1) + 1
      array_of_starts(0)   = stride(i,1) - 1
      call MPI_Type_create_subarray(ndims, ext, array_of_subsizes, array_of_starts, & 
                                    MPI_ORDER_FORTRAN, MPI_R_K, types(i), err)
      err = err - MPI_SUCCESS
      FD4_L3_ASSERT(err/=0,'MPI_Type_create_subarray failed')
      displ(i) = 0_MPI_ADDRESS_KIND
      blens(i) = 1
    end do

    if(count>1) then
      ! more than one subsection: combine them with MPI_Type_create_struct
      call MPI_Type_create_struct(count, blens, displ, types, sdata_type, err)
      err = err - MPI_SUCCESS
      FD4_L3_ASSERT(err/=0,'MPI_Type_create_struct failed')
      do i=1,count
        ! free the intermediate types
        call MPI_Type_free(types(i), err)
        err = err - MPI_SUCCESS
        FD4_L3_ASSERT(err/=0,'MPI_Type_free failed')
      end do
    else
      ! only on subsection
      sdata_type = types(1)    
    end if


!FD4_NO_MPI_TYPE_CREATE_SUBARRAY
#else

    !! using deprecated MPI_LB and MPI_UB

    count = MAX( (sbnd(3,2)-sbnd(3,1)+1) * (sbnd(2,2)-sbnd(2,1)+1), ubound(stride,1) ) + 2
    call fd4_comm_alloc_helpers(count)

    count = 0

    !! build type for dimension 0 considering stride

    count = count + 1
    types(count) = MPI_LB
    displ(count) = 0_MPI_ADDRESS_KIND
    blens(count) = 1
    do i=1,ubound(stride,1) 
      count = count + 1
      types(count) = MPI_R_K
      displ(count) = INT( (stride(i,1) - 1) * MPI_R_K_size, MPI_ADDRESS_KIND)
      blens(count) = stride(i,2) - stride(i,1) + 1
    end do
    count = count + 1
    types(count) = MPI_UB
    displ(count) = INT( ext(0) * MPI_R_K_size, MPI_ADDRESS_KIND)
    blens(count) = 1

    call MPI_Type_create_struct(count, blens, displ, types, cell_type, err)
    err = err - MPI_SUCCESS
    FD4_L3_ASSERT(err/=0,'MPI_Type_create_struct failed')


    !! build the type for the spatial dimension considering sbnd using the cell_type
    
    count = 0

    count = count + 1
    types(count) = MPI_LB
    displ(count) = 0_MPI_ADDRESS_KIND
    blens(count) = 1
    do k=sbnd(3,1),sbnd(3,2)
      do j=sbnd(2,1),sbnd(2,2)
        count = count + 1
        types(count) = cell_type
        displ(count) = INT( ext(0) * ( (sbnd(1,1)-1) + ext(1) * ( (j-1) + ext(2)*(k-1) ) ) * MPI_R_K_size , MPI_ADDRESS_KIND)
        blens(count) = sbnd(1,2)-sbnd(1,1)+1
      end do
    end do
    count = count + 1 ! UB
    types(count) = MPI_UB
    displ(count) = INT( PRODUCT(ext(0:3)) * MPI_R_K_size, MPI_ADDRESS_KIND)
    blens(count) = 1

    call MPI_Type_create_struct(count, blens, displ, types, sdata_type, err)
    err = err - MPI_SUCCESS
    FD4_L3_ASSERT(err/=0,'MPI_Type_create_struct failed')


    ! free the cell type, we only needed it for building the sdata_type
    call MPI_Type_free(cell_type, err)
    err = err - MPI_SUCCESS
    FD4_L3_ASSERT(err/=0,'MPI_Type_free failed')
    

!FD4_NO_MPI_TYPE_CREATE_SUBARRAY
#endif
    
    err = 0    
  
  end subroutine fd4_comm_build_array_type



  !! Allocate module variables blens, types, displ.
  !!
  !! Internal routine.
  subroutine fd4_comm_alloc_helpers(len)
    integer, intent(in) :: len
    
    if(allocated(blens)) then
      if(size(blens)>=len) return
      deallocate(blens)
      deallocate(types)
      deallocate(displ)    
    end if
    
    allocate(blens(len))
    allocate(types(len))
    allocate(displ(len))
  
  end subroutine fd4_comm_alloc_helpers

end module fd4_comm_mod
