! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010-2012 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!                 FD4 exchange of ghost blocks                 !
!                                                              !
!--------------------------------------------------------------!

#include "fd4.h"

!! Exchange of ghost blocks.
!!
!! @see fd4_comm_mod
!! @see fd4_neigh_mod
!! @see rbtree_fd4_neigh_mod
!! @author Matthias Lieber
module fd4_ghostcomm_mod

  use fd4_globaldef_mod
  use fd4_vartab_mod
  use fd4_block_mod
  use fd4_comm_mod
  use fd4_domain_mod
  use fd4_neigh_mod
  use fd4_mpitype_mod
  use fd4_mpi_mod
  use fd4_ghostlist_mod
  use fd4_stat_mod
  use rbtree_fd4_neigh_mod
  use stack_mod
  use mrgrnk_mod

  USE_MPI
  
  implicit none
  private
  
  INCLUDE_MPIFH
  
  public :: fd4_ghostblock_types, fd4_ghostcomm, fd4_domain, fd4_neigh
  public :: fd4_ghostcomm_exch, fd4_ghostcomm_create, fd4_ghostcomm_delete

  !! Save MPI datatypes for ghost communication of a block with specific size
  type fd4_ghostblock_types
    integer :: btypes(3,2)          !! block types for comm. in (dim,dir)
    integer :: gtypes(3)            !! ghost block types for comm. in (dim)
  end type fd4_ghostblock_types
  

  !! Type describing a ghost communication relation within a domain
  !! for a selection of variables and timesteps.
  !!
  type fd4_ghostcomm
    type(fd4_domain), pointer :: domain => null()     !! the domain
    type(fd4_ghostblock_types), pointer :: gbt(:,:,:) !! (ghost) block types for each existing spatial block size
    type(rbtree_fd4_neigh) :: neighstore              !! MPI types for each current neighbor are stored in a rbtree
    integer(i8k) :: partition_id                      !! unique identifier of the current partition (counter)
    logical :: exchdim(3)                             !! flag whether to exchange ghosts or not for each spatial dimension
    integer(i_k) :: stat_bytes                        !! index of the total ghostcomm size in the statistics object
  end type fd4_ghostcomm

  
contains
  
  include 'framework/fd4_domain.incf'
  include 'util/stack.incf'


  !! Create a new ghost communication relation with specified fields.
  !!
  !! The default value of $opt_exchdim is to exchange all dimensions for all variables.
  !! Valid values for $opt_exchdim are $FD4_X, $FD4_Y, $FD4_Z, $FD4_XY, $FD4_XZ, $FD4_YZ, and $FD4_XYZ.
  !! 
  !! *TODO*:
  !!  - input parameter checking
  !!  - allocate stat checking
  subroutine fd4_ghostcomm_create(gc, domain, nfields, idx, st, err, opt_exchdim)
    type(fd4_ghostcomm), intent(inout)       :: gc            !! the ghost communication description
    type(fd4_domain), intent(inout), target  :: domain        !! the domain
    integer(i_k), intent(in)                 :: nfields       !! number of fields in $idx and $st
    integer(i_k), intent(in)                 :: idx(nfields)  !! variable indexes
    integer(i_k), intent(in)                 :: st(nfields)   !! time steps
    integer(i_k), intent(out)                :: err           !! error status: 0...ok
    !! definition of spatial dimensions for ghost exchange for each variable
    integer(i_k), intent(in),optional        :: opt_exchdim(nfields)
    
    integer(i_k) :: i, j, k
    integer(i_k) :: dim, ext(0:3), sbnd(3,2)
    integer(i_k), pointer :: bnd(:,:)
    integer(i_k), pointer :: stride(:,:)    ! stride within $block%sdata describing the selected variables
    integer(i_k) :: ifld, nfld
    
    err = -1
    
    !TODO: input parameter checking
    
    allocate(stride(nfields,2))

    gc%domain => domain
    
    gc%partition_id = -1
    
    gc%exchdim(1:3) = .true.
    if(present(opt_exchdim)) then
     gc%exchdim(1) = ( MAXVAL(IAND(opt_exchdim,FD4_X))>0 )
     gc%exchdim(2) = ( MAXVAL(IAND(opt_exchdim,FD4_Y))>0 )
     gc%exchdim(3) = ( MAXVAL(IAND(opt_exchdim,FD4_Z))>0 )
    end if
    
    !! create MPI datatypes for ghost block communication
    
    bnd => domain%mmbsz  ! min and max block size in domain
    
    !print *,'min: ',bnd(:,1)
    !print *,'max: ',bnd(:,2)
    
    ! allocate the index array for each possible block size
    allocate( gc%gbt( bnd(1,1):bnd(1,2),bnd(2,1):bnd(2,2),bnd(3,1):bnd(3,2) ) )
    
    ext(0) = domain%varinfo%field_len(0)
    
    ! loop over all possible block sizes
    !TODO: if specific block sizes where given, some block sizes may be unused
    do k=bnd(3,1),bnd(3,2)
      do j=bnd(2,1),bnd(2,2)
        do i=bnd(1,1),bnd(1,2)
          ! loop over the 3 spatial dimensions
          do dim=1,3
            if(domain%nghosts(dim)<=0) cycle
            ! cycle the loop if no exchange in dimension dim requested
            if(.not. gc%exchdim(dim)) cycle            
            ! create stride for variable/timestep selection,
            if(present(opt_exchdim)) then
              ! check if current dimension is required (for each single variable)
              nfld = 0
              do ifld=1,nfields
                if( IAND(opt_exchdim(ifld), FD4_DIM(dim)) /= 0 ) then
                  nfld = nfld + 1
                  stride(nfld,1:2) = domain%varinfo%fields(idx(ifld),st(ifld),1:2)
                end if
              end do
            else
              ! all dimensions required
              do ifld=1,nfields
                stride(ifld,1:2) = domain%varinfo%fields(idx(ifld),st(ifld),1:2)
              end do
              nfld = nfields
            end if
            ext(3) = k
            ext(2) = j
            ext(1) = i
            sbnd(1:3,1) = 1
            sbnd(1:3,2) = ext(1:3)
            !! construct the type for a block of size ext(:)
            ! upper face
            sbnd(dim,1) = ext(dim) - domain%nghosts(dim) + 1
            sbnd(dim,2) = ext(dim)
            call fd4_comm_build_array_type(ext, stride(1:nfld,1:2), sbnd, gc%gbt(i,j,k)%btypes(dim,2), err)
            FD4_L1_ASSERT(err/=0,'fd4_comm_build_array_type failed')
            ! lower face
            sbnd(dim,1) = 1
            sbnd(dim,2) = domain%nghosts(dim)
            call fd4_comm_build_array_type(ext, stride(1:nfld,1:2), sbnd, gc%gbt(i,j,k)%btypes(dim,1), err)
            FD4_L1_ASSERT(err/=0,'fd4_comm_build_array_type failed')
            !! construct the type for a ghost block of size ext(:)
            ext(dim) = domain%nghosts(dim)
            ! sbnd same values as above
            call fd4_comm_build_array_type(ext, stride(1:nfld,1:2), sbnd, gc%gbt(i,j,k)%gtypes(dim), err)
            FD4_L1_ASSERT(err/=0,'fd4_comm_build_array_type failed')
          end do
          ! end of loop over the 3 spatial dimensions
        end do
      end do
    end do
    
    deallocate(stride)
    
    call fd4_stat_add(domain%stat,'Ghostcomm Bytes',gc%stat_bytes)

    err = 0    
  
  end subroutine fd4_ghostcomm_create
  
  

  !! Delete ghost communication relation.
  !!
  subroutine fd4_ghostcomm_delete(gc)
    type(fd4_ghostcomm), intent(inout):: gc            !! the ghost communication description
    
    integer(i_k) :: i, j, k
    integer(i_k) :: dim, dir, err, bnd(3,2)
    
    ! deallocate neighbor process data including data types
    call rbtree_fd4_neigh_clear(gc%neighstore,fd4_neigh_deallocate_ptr)

    ! delete datatypes of blocks
    bnd(:,1) = lbound(gc%gbt)
    bnd(:,2) = ubound(gc%gbt)
    do k=bnd(3,1),bnd(3,2)
      do j=bnd(2,1),bnd(2,2)
        do i=bnd(1,1),bnd(1,2)
          do dim=1,3
            if(gc%domain%nghosts(dim)<=0) cycle
            ! cycle the loop if no exchange in dimension dim requested
            if(.not. gc%exchdim(dim)) cycle            
            do dir=1,2
              call MPI_Type_free(gc%gbt(i,j,k)%btypes(dim,dir), err)
              err = err - MPI_SUCCESS
              FD4_L3_ASSERT(err/=0,'MPI_Type_free failed')
            end do
            call MPI_Type_free(gc%gbt(i,j,k)%gtypes(dim), err)
            err = err - MPI_SUCCESS
            FD4_L3_ASSERT(err/=0,'MPI_Type_free failed')
          end do
        end do
      end do
    end do

    deallocate(gc%gbt)

    ! reset contents
    gc%domain => null()
    gc%partition_id = -1
    
  end subroutine fd4_ghostcomm_delete
    
  

  !! Exchange of ghost blocks.
  !!
  !!  - if partitioning has changed: rebuild the MPI datatypes for the ghost block communication
  !!  - foreach neighbor process:
  !!    - start an MPI_Irecv for all the ghost blocks
  !!    - start an MPI_Isend for all the parent blocks ghost zones
  !!  - MPI_Waitall for all requests
  !!
  subroutine fd4_ghostcomm_exch(gc, err)
    type(fd4_ghostcomm), intent(inout), target :: gc            !! the ghost communication description
    integer(i_k), intent(out)                  :: err           !! error status: 0...ok
    
    type(fd4_block), pointer     :: gb, b
    integer(i_k) :: hidx, ghidx, pos(3), dim, dir, i
    integer(i_k) :: dest
    integer(i8k) :: sendtag, recvtag
    integer(i8k), parameter :: i33 = 2_i8k**32 ! bit #33 of a 64 bit number
    integer(i8k), parameter :: magic(3,2) = reshape ((/ 0_i8k, i33, 2*i33, 3*i33, 4*i33, 5*i33 /), (/ 3, 2 /))
    integer :: request
    type(fd4_mpi), pointer :: mpi
    type(fd4_neigh), pointer :: n, deleted
    integer(MPI_ADDRESS_KIND) :: disp
    logical :: is_min
    type(rbtree_fd4_neigh_iter) :: niter
    
    err = -1

    mpi => gc%domain%mpi

    mpi%requests%pos = 0

    !! check if domain's partitioning has changed (only on this process) since last call to fd4_ghostcomm_exch with this gc
    if(gc%partition_id /= gc%domain%partition_id) then
    
      ! free the mpi types of each fd4_neigh
      ! but keep the fd4_neighs allocated, because they might get reused
      n => rbtree_fd4_neigh_iter_first(niter, gc%neighstore)
      do while(associated(n))
        call fd4_mpitype_free(n%mpitype(FD4_NEIGH_SEND), .false., .false.)
        call fd4_mpitype_free(n%mpitype(FD4_NEIGH_RECV), .false., .false.)
        n%changed = .false.
        n => rbtree_fd4_neigh_iter_next(niter)
      end do

      !! loop over ghostlist
      gb => gc%domain%glist(FD4_GHOST_CMM)%start
      do while(associated(gb))
        dim = gb%pos(3)  ! spatial dimension: 1, 2, or 3
        
        ! is exchange in dimension dim requested?
        if(gc%exchdim(dim)) then

          dir = gb%pos(2)  ! direction: 1 or 2
          b => gb%neigh(2,1)%l   ! the block within the local domain ("parent" of the ghost)

          FD4_L3_ASSERT(gc%domain%nghosts(dim)==0,'Malicious ghost block')

          ! ghosts's position
          pos(1:3) = b%pos(1:3)
          pos(dim) = pos(dim) + fd4_incdec(dir)
          ! set periodic boundary
          ! (at-boundary-but-no-periodic-boundaries-case not possible here)
          if( pos(dim) < 1 ) then
            pos(dim) = gc%domain%blnum(dim)
          else if( pos(dim) > gc%domain%blnum(dim) ) then
            pos(dim) = 1
          end if

          call fd4_domain_c2i(b%pos, hidx)      !  parent block idx
          call fd4_domain_c2i(pos, ghidx)       !  ghost blocks idx (i.e. remote block idx)

          ! create sorting indices for fd4_mpitype_add
          ! [32 bit idx of senderblock] + [encoded send direction using only the high bits]
          sendtag =  INT( hidx,i8k) + magic(dim,  dir)
          recvtag =  INT(ghidx,i8k) + magic(dim,3-dir)
          !dest = gc%domain%blkmap(pos(1),pos(2),pos(3),gc%domain%blkmap_now)
          dest = gb%sempty

          FD4_L3_ASSERT(dest==mpi%rank,'Oops! Sending ghost block to myself...')

          ! check if a neigh type for this rank already exists, if not, create
          n => rbtree_fd4_neigh_get(gc%neighstore, dest)
          if(.not. associated(n)) then
            allocate(n, stat=err)
            FD4_L3_ASSERT(err/=0,'allocate neigh')
            call rbtree_fd4_neigh_insert (gc%neighstore, dest, n)
          end if
          n%changed = .true.
          
          !! get parameters for building recv type
          call MPI_Get_address( gb%sdata(0)%l(1,1,1,1), disp, err)
          i = gc%gbt(b%ext(1),b%ext(2),b%ext(3))%gtypes(dim) ! type for the block
          call fd4_mpitype_add( n%mpitype(FD4_NEIGH_RECV), recvtag, 1, disp, i, is_min)
          if(is_min) n%minptr(FD4_NEIGH_RECV)%l => gb%sdata(0)%l(1,1,1,1)

          !! get parameters for building send type
          call MPI_Get_address( b%sdata(0)%l(1,1,1,1), disp, err)
          i = gc%gbt(b%ext(1),b%ext(2),b%ext(3))%btypes(dim,dir) ! type for the block
          call fd4_mpitype_add( n%mpitype(FD4_NEIGH_SEND), sendtag, 1, disp, i, is_min)
          if(is_min) n%minptr(FD4_NEIGH_SEND)%l => b%sdata(0)%l(1,1,1,1)

#if ( FD4_VERBOSE_LEVEL >= 5 )
           write(fd4_out,'(A,A)') FD4_LINEHD,'exch ghost: idx: '//i2s(hidx)//i2s(ghidx)//  &
                                             ' tag: '//l2s(sendtag)//l2s(recvtag)//' dest: '//i2s(dest)
#endif

        end if ! is exchange in dimension dim requested?
        
        gb => gb%neigh(1,2)%l  ! iterate to next ghost block
      end do
      
      ! remove fd4_neighs which have not been reused in the ghostlist loop
      dest = -1
      n => rbtree_fd4_neigh_iter_first(niter, gc%neighstore)
      do while(associated(n))
        if(.not. n%changed) dest = niter%cur%key
        n => rbtree_fd4_neigh_iter_next(niter)
        if(dest>=0) then
          ! delete previous item and fix iterator
          deleted => rbtree_fd4_neigh_delete(gc%neighstore, dest, niter)
          call fd4_neigh_deallocate_ptr(deleted)
          dest = -1
        end if
      end do

      gc%partition_id = gc%domain%partition_id

      FD4_L4_INFO('Updated ghostcomm type to '//l2s(gc%domain%partition_id))
      FD4_L4_INFO('Ghostlist size '//i2s(gc%domain%glist(FD4_GHOST_CMM)%length))
      FD4_L4_INFO('Neighstore size '//i2s(gc%neighstore%length))

    end if ! partitioning has changed

    
    !! send
    !! loop over neighbor processes
    n => rbtree_fd4_neigh_iter_first(niter, gc%neighstore)
    do while(associated(n))
      if(n%mpitype(FD4_NEIGH_SEND)%count>0) then
        dest = niter%cur%key

        if( n%mpitype(FD4_NEIGH_SEND)%datatype == MPI_DATATYPE_NULL ) then
          ! we need to construct datatype for all ghost communication with rank $dest

          call fd4_mpitype_commit(n%mpitype(FD4_NEIGH_SEND), err)
          FD4_L3_ASSERT(err/=0,'fd4_mpitype_commit (FD4_NEIGH_SEND) failed')

!#define FD4_MPI_TYPE_SIZE
#ifdef FD4_MPI_TYPE_SIZE
          call MPI_Type_size(n%mpitype(FD4_NEIGH_SEND)%datatype, i, err)
          call MPI_Type_extent(n%mpitype(FD4_NEIGH_SEND)%datatype, hidx, err)
          FD4_L4_INFO('FD4_NEIGH_SEND size & extent (bytes): '//i2s(i)//i2s(hidx))
#endif

        end if

        ! finally send the message
        call MPI_Isend(n%minptr( FD4_NEIGH_SEND)%l, 1, n%mpitype(FD4_NEIGH_SEND)%datatype, dest, &
                                 n%mpitype(FD4_NEIGH_SEND)%count, mpi%comm_gc, request, err )
        err = err - MPI_SUCCESS
        FD4_L3_ASSERT(err/=0,'MPI_Isend failed')
        call stack_push(mpi%requests, request)

      end if
      n => rbtree_fd4_neigh_iter_next(niter)
    end do
      
    !! receive
    !! loop over neighbor processes
    n => rbtree_fd4_neigh_iter_first(niter, gc%neighstore)
    do while(associated(n))
      if(n%mpitype(FD4_NEIGH_RECV)%count>0) then
        dest = niter%cur%key

        if( n%mpitype(FD4_NEIGH_RECV)%datatype == MPI_DATATYPE_NULL ) then
          ! we need to construct datatype for all ghost communication with rank $dest

          call fd4_mpitype_commit(n%mpitype(FD4_NEIGH_RECV), err)
          FD4_L3_ASSERT(err/=0,'fd4_mpitype_commit (FD4_NEIGH_RECV) failed')

#ifdef FD4_MPI_TYPE_SIZE
          call MPI_Type_size(n%mpitype(FD4_NEIGH_RECV)%datatype, i, err)
          call MPI_Type_extent(n%mpitype(FD4_NEIGH_RECV)%datatype, hidx, err)
          FD4_L4_INFO('FD4_NEIGH_RECV size & extent (bytes): '//i2s(i)//i2s(hidx))
#endif

        end if

        if(gc%domain%stat_enabled) then
          call MPI_Type_size(n%mpitype(FD4_NEIGH_RECV)%datatype, i, err)
          gc%domain%stat%values(gc%stat_bytes) = gc%domain%stat%values(gc%stat_bytes) + INT(i,i8k)
        end if

        ! finally receive the message
        call MPI_Irecv(n%minptr( FD4_NEIGH_RECV)%l, 1, n%mpitype(FD4_NEIGH_RECV)%datatype, dest, &
                                 n%mpitype(FD4_NEIGH_RECV)%count, mpi%comm_gc, request, err )
        err = err - MPI_SUCCESS
        FD4_L3_ASSERT(err/=0,'MPI_Irecv failed')
        call stack_push(mpi%requests, request)

      end if
      n => rbtree_fd4_neigh_iter_next(niter)
    end do


    !! wait for the communication to complete
    
    if(mpi%requests%pos>0) then
      call MPI_Waitall(mpi%requests%pos, mpi%requests%a, MPI_STATUSES_IGNORE, err)
      err = err - MPI_SUCCESS
      FD4_L3_ASSERT(err/=0,'MPI_Waitall failed')
    end if    



    err = 0    
    
  end subroutine fd4_ghostcomm_exch



end module fd4_ghostcomm_mod
