! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010-2012 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!                      Iterator over blocks                    !
!                                                              !
!--------------------------------------------------------------!

#include "fd4.h"

!! Module for an iterator over blocks.
!!
!! Simple usage example without ghost cells:
!!>  call fd4_iter_init(domain, iter)
!!   do while(associated(iter%cur))
!!     ! iter%cur%bnd, iter%cur%pos contain bounds and block position
!!     ! get one field
!!     call fd4_iter_get(iter,field_idx,array_ext,array,block_ext)
!!     ! loop over this block's grid cells
!!     ! integer(i_k) :: block_ext(0:3) contains actual block size
!!     do k=1,block_ext(3)
!!       do j=1,block_ext(2)
!!         do i=1,block_ext(1)
!!           ! do some computations with the data at this spatial position
!!           call compute(array(1:block_ext(0),i,j,k))
!!         end do
!!       end do
!!     end do
!!     ! put data back to the framework
!!     call fd4_iter_put(iter,field_idx,array_ext,array)
!!     ! iterate to next block
!!     call fd4_iter_next(iter)
!!<  end do
!!
!! *TODO*: 
!!  - $fd4_iter_bext - get only bext without get or put
!!  - get/put for 1-bin-variables (using 3D array instead 4D)
!!  - several variants of get/put-routines
!!  - ghost block support
!!  - dynamic data support
!! 
!!
!! @see fd4_block_mod
!! @author Matthias Lieber
module fd4_iter_mod

  use timing_mod
  use fd4_globaldef_mod
  use fd4_vartab_mod
  use fd4_block_mod
  use fd4_domain_mod
  use rbtree_fd4_block_mod
  
  implicit none
  private
  
  ! public types
  public :: fd4_domain, fd4_block, rbtree_fd4_block_iter
  public :: fd4_iter

  ! public methods
  public :: fd4_iter_init
  public :: fd4_iter_next
  public :: fd4_iter_valid
  public :: fd4_iter_offset
  public :: fd4_iter_block_extent
  public :: fd4_iter_get
  public :: fd4_iter_put
  public :: fd4_iter_get_ghost
  public :: fd4_iter_get_facevar
  public :: fd4_iter_put_ghost
  public :: fd4_iter_get_ghost_multi
  public :: fd4_iter_empty
  public :: fd4_iter_start_clock
  public :: fd4_iter_stop_clock
  
  
  !! This is the block iterator type.
  !! You can read the components of the type for fast access, but do not modify them!
  type fd4_iter
  
    type(fd4_domain), pointer :: domain => null()  !! the domain which owns the blocks to iterate
    type(fd4_block),  pointer :: cur => null()     !! current block
    type(rbtree_fd4_block_iter) :: biter       !! iterator for block data structure
    integer(i8k) :: time_start = 0                 !! start time of calculation on current block
    
  end type fd4_iter

contains  


  !! Set iterator to first block of a given domain.
  !!
  !! @see fd4_domain_mod
  subroutine fd4_iter_init(domain, iter)
    implicit none  
    type(fd4_domain), intent(in), target :: domain !! the domain
    type(fd4_iter),   intent(inout)      :: iter   !! the iterator
    iter%domain  => domain
    FD4_L3_ASSERT(.not.iter%domain%init,'domain not initialized')
    iter%cur => rbtree_fd4_block_iter_first(iter%biter, iter%domain%blocks)
  end subroutine fd4_iter_init


  !! Iterate to next block.
  !! 
  !! *CAUTION*: This routine will intentionally cause a segmentation violation
  !! if the iterator is not valid and verbose level < 3!
  subroutine fd4_iter_next(iter)
    implicit none  
    type(fd4_iter), intent(inout) :: iter        !! the iterator
    FD4_L3_ASSERT(.not.associated(iter%cur),'invalid iterator')
    iter%cur => rbtree_fd4_block_iter_next(iter%biter)
  end subroutine fd4_iter_next


  !! Check if iterator points to block.
  !!
  !! Currently, this is equal to
  !!v associated(iter%cur)
  logical function fd4_iter_valid(iter)
    implicit none  
    type(fd4_iter), intent(in) :: iter !! the iterator
    fd4_iter_valid = associated(iter%cur)
  end function fd4_iter_valid


  !! Get spatial offset from block local coordinates to global coordinates.
  !!
  !! Block local coordinates are starting with 1. This subroutine returns
  !! the offset which must be added to the local coordinates to retrieve
  !! the global domain coordinates. E.g.:
  !!v global_x = block_x + offset(1)
  !!
  !! Currently, the value of $offset is:
  !!v iter%cur%bnd(:,1) - 1
  !!
  !! *CAUTION*: This routine will intentionally cause a segmentation violation
  !! if the iterator is not valid and verbose level < 3!
  subroutine fd4_iter_offset(iter, offset)
    implicit none  
    type(fd4_iter), intent(in)  :: iter      !! the iterator
    integer(i_k),   intent(out) :: offset(3) !! local-to-global offset
    FD4_L3_ASSERT(.not.associated(iter%cur),'invalid iterator')
    offset(:) = iter%cur%bnd(:,1) - 1
  end subroutine fd4_iter_offset



  !! Get spatial extent of field $idx within current block.
  !!
  !! Returns the same $bext as the $fd4_iter_get-Routines do.
  subroutine fd4_iter_block_extent(iter, idx, ghosts, bext)
    implicit none  
    type(fd4_iter), intent(in)    :: iter       !! the iterator
    integer(i_k),   intent(in)    :: idx        !! field index in vartab
    logical,        intent(in)    :: ghosts     !! add ghost cells to extent? (cell-centered vars only)
    integer(i_k),   intent(out)   :: bext(0:3)  !! actual extents of the block
    bext(0)   = iter%domain%vartab(idx)%nbins
    bext(1:3) = iter%cur%ext(:)
    if(iter%domain%vartab(idx)%facevar > 0) bext(iter%domain%vartab(idx)%facevar) = bext(iter%domain%vartab(idx)%facevar) + 1
    if(ghosts) bext(1:3) = bext(1:3) + 2 * iter%domain%nghosts(1:3)
  end subroutine fd4_iter_block_extent
  
  
  !! Safe the current time as the starting time of calculations for the current block.
  !!
  subroutine fd4_iter_start_clock(iter)
    implicit none  
    type(fd4_iter), intent(inout)    :: iter       !! the iterator
    call gettime(iter%time_start, timer=iter%domain%timer)
  end subroutine fd4_iter_start_clock


  !! Get the ending time of calculations for the current block.
  !!
  !! Sets the block's weight to the time interval between calls to $fd4_iter_start_clock and $fd4_iter_stop_clock.
  !! The argument $fact enables to add the time interval to the block weight ($fact=1.0) or
  !! replacing the value ($fact=0.0).
  !! More generally, $fact is multiplied to the current weight before adding the time interval.
  !!
  subroutine fd4_iter_stop_clock(iter, fact)
    implicit none  
    type(fd4_iter), intent(inout) :: iter    !! the iterator
    real(r4k),      intent(in) :: fact       !! add the time to block weight (1.0) or replace (0.0)
    integer(i8k) :: time_end
    call gettime(time_end, timer=iter%domain%timer)
    iter%cur%weight = fact * iter%cur%weight + REAL(time_end - iter%time_start, r4k)    
  end subroutine fd4_iter_stop_clock


  !! Get data from the current block.
  !!
  !! This is the order of the indices in $aext, $bext, and $array: (bins, x, y, z)
  !!
  !! *CAUTION*: Error checking will be enabled only if verbose level >= 3.
  !! If not: *Absolutely no checking for error conditions will be carried through*!
  !! This routine will intentionally cause a segmentation violation if the iterator
  !! is not valid, the field $idx does not exist, or the array is to small!
  !!
  !! *TODO*:
  !!  - boundary from neighbor blocks: extra functions for *diagonal* stencil, *trigonal* stencil
  !!  - let user choose which boundaries are needed (e.g. x,y only for horizontal advection)
  !!  - ghost blocks
  !!  - get multiple (all) fields of a block at once
  !!  - selection of spectral and spatial parts of the block?
  !!  - what happens, if dynamic field is not allocated? put null value to array?
  subroutine fd4_iter_get(iter, idx, st, aext, array, opt_bext)
    implicit none  
    type(fd4_iter), intent(in)    :: iter       !! the iterator
    integer(i_k),   intent(in)    :: idx        !! field index in vartab
    integer(i_k),   intent(in)    :: st         !! time step index
    integer(i_k),   intent(in)    :: aext(0:3)  !! extents of the buffer array
    real(r_k),      intent(inout) :: array(aext(0),aext(1),aext(2),aext(3))    !! buffer array to write data
    integer(i_k),   intent(out)   :: opt_bext(0:3)  !! actual extents of the block written to the buffer array
    optional :: opt_bext
    
    integer(i_k) :: bext(0:3)        !! block extent, inclusive bins

    bext(0)   = iter%domain%vartab(idx)%nbins
    bext(1:3) = iter%cur%ext(:)

#if ( FD4_VERBOSE_LEVEL >= 3 )
      ! error checks
      FD4_L3_ASSERT(.not.associated(iter%cur),'invalid iterator')
      if(idx<0 .or. idx>ubound(iter%domain%vartab,1)) then
        FD4_L1_ERROR('field idx not valid')
      end if
      if(st<0 .or. st>iter%domain%vartab(idx)%nsteps) then
        FD4_L1_ERROR('time step not valid')
      end if
      ! TODO: might change in future
      if(.not.associated(iter%cur%fields(idx,st)%l)) then
        FD4_L1_ERROR('field not allocated')
      end if
      if(iter%domain%vartab(idx)%facevar/=0) then
        FD4_L1_ERROR('no cell-centered variable')
      end if
      if(bext(0)>aext(0) .or. bext(1)>aext(1) .or. bext(2)>aext(2) .or. bext(3)>aext(3)) then
        FD4_L1_ERROR('array too small')
      end if
#endif


    array(1:bext(0),1:bext(1),1:bext(2),1:bext(3)) = iter%cur%fields(idx,st)%l(:,:,:,:)
    
    if(present(opt_bext)) opt_bext = bext

    
  end subroutine fd4_iter_get



  !! Put data to the current block.
  !!
  !! This is the order of the indices in $aext, $bext, and $array: (bins, x, y, z)
  !!
  !! *CAUTION*: Error checking will be enabled only if verbose level >= 3.
  !! If not: *Absolutely no checking for error conditions will be carried through*!
  !! This routine will intentionally cause a segmentation violation if the iterator
  !! is not valid, the field $idx does not exist, or the array is to small!
  !!
  !! *TODO*:
  !!  - see $fd4_iter_get
  !!  - optionally call $fd4_block_empty? (like in $fd4_iter_put_ghost)
  !!
  subroutine fd4_iter_put(iter, idx, st, aext, array)
    implicit none  
    type(fd4_iter), intent(inout) :: iter       !! the iterator
    integer(i_k),   intent(in)    :: idx        !! field index in vartab
    integer(i_k),   intent(in)    :: st         !! time step index
    integer(i_k),   intent(in)    :: aext(0:3)  !! extents of the buffer array
    real(r_k),      intent(in)    :: array(aext(0),aext(1),aext(2),aext(3))    !! buffer array to read data
    integer(i_k) :: bext(0:3)    !! actual extents of the block written to the buffer array

    bext(0)   = iter%domain%vartab(idx)%nbins
    bext(1:3) = iter%cur%ext(:)

#if ( FD4_VERBOSE_LEVEL >= 3 )
      ! error checks
      FD4_L3_ASSERT(.not.associated(iter%cur),'invalid iterator')
      if(idx<0 .or. idx>ubound(iter%domain%vartab,1)) then
        FD4_L1_ERROR('field idx not valid')
      end if
      if(st<0 .or. st>iter%domain%vartab(idx)%nsteps) then
        FD4_L1_ERROR('time step not valid')
      end if
      ! TODO: might change in future
      if(.not.associated(iter%cur%fields(idx,st)%l)) then
        FD4_L1_ERROR('field not allocated')
      end if
      if(iter%domain%vartab(idx)%facevar/=0) then
        FD4_L1_ERROR('no cell-centered variable')
      end if
      if(bext(0)>aext(0) .or. bext(1)>aext(1) .or. bext(2)>aext(2) .or. bext(3)>aext(3)) then
        FD4_L1_ERROR('array too small')
      end if
#endif


    iter%cur%fields(idx,st)%l(:,:,:,:) = array(1:bext(0),1:bext(1),1:bext(2),1:bext(3))

    
  end subroutine fd4_iter_put




  !! Get data from the current block with boundary.
  !!
  !! Sets values at boundaries without neighbor block $vnull.
  !! The order of the indices in $aext, $opt_bext, and $array is: (bins, x, y, z)
  !!
  !! *CAUTION*: Error checking will be enabled only if verbose level >= 3.
  !! If not: *Absolutely no checking for error conditions will be carried through*!
  !! This routine will intentionally cause a segmentation violation if the iterator
  !! is not valid, the field $idx does not exist, or the array is to small!
  !!
  !! *TODO*:
  !!  - see $fd4_iter_get
  !!  - return error status?
  !!
  subroutine fd4_iter_get_ghost(iter, idx, st, aext, array, opt_bext)
    implicit none  
    type(fd4_iter), intent(in)    :: iter       !! the iterator
    integer(i_k),   intent(in)    :: idx        !! field index in vartab
    integer(i_k),   intent(in)    :: st         !! time step index
    integer(i_k),   intent(in)    :: aext(0:3)  !! extents of the buffer array
    real(r_k),      intent(inout) :: array(aext(0),aext(1),aext(2),aext(3))    !! buffer array to write data
    integer(i_k),   intent(out)   :: opt_bext(0:3)  !! actual extents of the block written to the buffer array
    optional :: opt_bext
    
    integer(i_k) :: ng(3)            !! = iter%domain%nghosts, number of ghosts cells
    integer(i_k) :: nbext(0:3)       !! upper bounds of a neighbor block array
    integer(i_k) :: bext(0:3)        !! block extent, inclusive bins
#ifdef FD4_DIAGONAL_HORIZONTAL
    type(fd4_block), pointer :: b
    target :: iter
#endif        

    ng = iter%domain%nghosts
    bext(0)   = iter%domain%vartab(idx)%nbins
    bext(1:3) = iter%cur%ext(:)

#if ( FD4_VERBOSE_LEVEL >= 3 )
      ! error checks
      FD4_L3_ASSERT(.not.associated(iter%cur),'invalid iterator')
      if(idx<0 .or. idx>ubound(iter%domain%vartab,1)) then
        FD4_L1_ERROR('field idx not valid')
      end if
      if(st<0 .or. st>iter%domain%vartab(idx)%nsteps) then
        FD4_L1_ERROR('time step not valid')
      end if
      ! TODO: might change in future
      if(.not.associated(iter%cur%fields(idx,st)%l)) then
        FD4_L1_ERROR('field not allocated')
      end if
      if(iter%domain%vartab(idx)%facevar/=0) then
        FD4_L1_ERROR('no cell-centered variable')
      end if
      if(bext(0)>aext(0) .or. bext(1)+2*ng(1)>aext(1) .or. bext(2)+2*ng(2)>aext(2) .or. bext(3)+2*ng(3)>aext(3)) then
        FD4_L1_ERROR('array too small')
      end if
#endif

    ! -- copy current block's data --
    array( 1:bext(0) , 1+ng(1):bext(1)+ng(1) , 1+ng(2):bext(2)+ng(2) , 1+ng(3):bext(3)+ng(3) ) = & 
         iter%cur%fields(idx,st)%l(:,:,:,:)

    
    ! -- copy neighbor blocks' data --

    ! TODO: if this field is dynamic, it might be that the neighbor block does not have allocated it!
    
    ! (-1, 0, 0)
    if(associated(iter%cur%neigh(1,1)%l)) then
      nbext = ubound( iter%cur%neigh(1,1)%l%fields(idx,st)%l )
      ! TODO: return error status?  
      FD4_L3_ASSERT(nbext(0)/=bext(0),'neigh(1,1) nbext(0)/=bext(0)')
      FD4_L3_ASSERT(nbext(2)/=bext(2),'neigh(1,1) nbext(2)/=bext(2)')
      FD4_L3_ASSERT(nbext(3)/=bext(3),'neigh(1,1) nbext(3)/=bext(3)')      
      array( 1:bext(0) , 1:ng(1) , 1+ng(2):bext(2)+ng(2) , 1+ng(3):bext(3)+ng(3) ) = & 
           iter%cur%neigh(1,1)%l%fields(idx,st)%l( : , nbext(1)-ng(1)+1:nbext(1) , : , :)
#ifdef FD4_DIAGONAL_HORIZONTAL
      b => iter%cur%neigh(1,1)%l

    ! (-1,-1, 0)
      if(associated(b%neigh(2,1)%l)) then
        nbext = ubound( b%neigh(2,1)%l%fields(idx,st)%l )
        FD4_L3_ASSERT(nbext(0)/=bext(0),'(-1,-1, 0) nbext(0)/=bext(0)')
        FD4_L3_ASSERT(nbext(3)/=bext(3),'(-1,-1, 0) nbext(3)/=bext(3)')      
        array( 1:bext(0) , 1:ng(1) , 1:ng(2) , 1+ng(3):bext(3)+ng(3) ) = & 
             b%neigh(2,1)%l%fields(idx,st)%l( : , nbext(1)-ng(1)+1:nbext(1) , nbext(2)-ng(2)+1:nbext(2) , : )
      else
        array( 1:bext(0) , 1:ng(1) , 1:ng(2) , 1+ng(3):bext(3)+ng(3) ) = & 
             iter%domain%vartab(idx)%vnull
      end if

    ! (-1, 1, 0)
      if(associated(b%neigh(2,2)%l)) then
        nbext = ubound( b%neigh(2,2)%l%fields(idx,st)%l )
        FD4_L3_ASSERT(nbext(0)/=bext(0),'(-1,-1, 0) nbext(0)/=bext(0)')
        FD4_L3_ASSERT(nbext(3)/=bext(3),'(-1,-1, 0) nbext(3)/=bext(3)')      
        array( 1:bext(0) , 1:ng(1) , bext(2)+ng(2)+1:bext(2)+2*ng(2) , 1+ng(3):bext(3)+ng(3) ) = & 
             b%neigh(2,2)%l%fields(idx,st)%l( : , nbext(1)-ng(1)+1:nbext(1) , 1:ng(2) , : )
      else
        array( 1:bext(0) , 1:ng(1) , bext(2)+ng(2)+1:bext(2)+2*ng(2) , 1+ng(3):bext(3)+ng(3) ) = & 
             iter%domain%vartab(idx)%vnull
      end if

#endif
    else
#ifdef FD4_DIAGONAL_HORIZONTAL
    ! (-1,-1, 0)
      array( 1:bext(0) , 1:ng(1) , 1:ng(2) , 1+ng(3):bext(3)+ng(3) ) = & 
           iter%domain%vartab(idx)%vnull
#endif
      array( 1:bext(0) , 1:ng(1) , 1+ng(2):bext(2)+ng(2) , 1+ng(3):bext(3)+ng(3) ) = &
           iter%domain%vartab(idx)%vnull
#ifdef FD4_DIAGONAL_HORIZONTAL
    ! (-1, 1, 0)
      array( 1:bext(0) , 1:ng(1) , bext(2)+ng(2)+1:bext(2)+2*ng(2) , 1+ng(3):bext(3)+ng(3) ) = & 
           iter%domain%vartab(idx)%vnull
#endif
    end if

    ! ( 1, 0, 0)
    if(associated(iter%cur%neigh(1,2)%l)) then
      nbext = ubound( iter%cur%neigh(1,2)%l%fields(idx,st)%l )
      ! TODO: return error status?  
      FD4_L3_ASSERT(nbext(0)/=bext(0),'neigh(1,2) nbext(0)/=bext(0)')
      FD4_L3_ASSERT(nbext(2)/=bext(2),'neigh(1,2) nbext(2)/=bext(2)')      
      FD4_L3_ASSERT(nbext(3)/=bext(3),'neigh(1,2) nbext(3)/=bext(3)')
      array( 1:bext(0) , bext(1)+ng(1)+1:bext(1)+2*ng(1) , 1+ng(2):bext(2)+ng(2) , 1+ng(3):bext(3)+ng(3) ) = & 
           iter%cur%neigh(1,2)%l%fields(idx,st)%l( : , 1:ng(1) , : , : )
#ifdef FD4_DIAGONAL_HORIZONTAL
      b => iter%cur%neigh(1,2)%l

    ! ( 1,-1, 0)
      if(associated(b%neigh(2,1)%l)) then
        nbext = ubound( b%neigh(2,1)%l%fields(idx,st)%l )
        FD4_L3_ASSERT(nbext(0)/=bext(0),'(-1,-1, 0) nbext(0)/=bext(0)')
        FD4_L3_ASSERT(nbext(3)/=bext(3),'(-1,-1, 0) nbext(3)/=bext(3)')      
        array( 1:bext(0) , bext(1)+ng(1)+1:bext(1)+2*ng(1) , 1:ng(2) , 1+ng(3):bext(3)+ng(3) ) = & 
             b%neigh(2,1)%l%fields(idx,st)%l( : , 1:ng(1) , nbext(2)-ng(2)+1:nbext(2) , : )
      else
        array( 1:bext(0) , bext(1)+ng(1)+1:bext(1)+2*ng(1) , 1:ng(2) , 1+ng(3):bext(3)+ng(3) ) = & 
             iter%domain%vartab(idx)%vnull
      end if

    ! ( 1, 1, 0)
      if(associated(b%neigh(2,2)%l)) then
        nbext = ubound( b%neigh(2,2)%l%fields(idx,st)%l )
        FD4_L3_ASSERT(nbext(0)/=bext(0),'(-1,-1, 0) nbext(0)/=bext(0)')
        FD4_L3_ASSERT(nbext(3)/=bext(3),'(-1,-1, 0) nbext(3)/=bext(3)')      
        array( 1:bext(0) , bext(1)+ng(1)+1:bext(1)+2*ng(1) , bext(2)+ng(2)+1:bext(2)+2*ng(2) , 1+ng(3):bext(3)+ng(3) ) = &
             b%neigh(2,2)%l%fields(idx,st)%l( : , 1:ng(1), 1:ng(2) , : )
      else
        array( 1:bext(0) , bext(1)+ng(1)+1:bext(1)+2*ng(1) , bext(2)+ng(2)+1:bext(2)+2*ng(2) , 1+ng(3):bext(3)+ng(3) ) = &
             iter%domain%vartab(idx)%vnull
      end if

#endif
    else
#ifdef FD4_DIAGONAL_HORIZONTAL
    ! ( 1,-1, 0)
      array( 1:bext(0) , bext(1)+ng(1)+1:bext(1)+2*ng(1) , 1:ng(2) , 1+ng(3):bext(3)+ng(3) ) = & 
           iter%domain%vartab(idx)%vnull
#endif
      array( 1:bext(0) , bext(1)+ng(1)+1:bext(1)+2*ng(1) , 1+ng(2):bext(2)+ng(2) , 1+ng(3):bext(3)+ng(3) ) = & 
           iter%domain%vartab(idx)%vnull
#ifdef FD4_DIAGONAL_HORIZONTAL
    ! ( 1, 1, 0)
      array( 1:bext(0) , bext(1)+ng(1)+1:bext(1)+2*ng(1) , bext(2)+ng(2)+1:bext(2)+2*ng(2) , 1+ng(3):bext(3)+ng(3) ) = &
           iter%domain%vartab(idx)%vnull
#endif
    end if
    
    ! ( 0,-1, 0)
    if(associated(iter%cur%neigh(2,1)%l)) then
      nbext = ubound( iter%cur%neigh(2,1)%l%fields(idx,st)%l )
      ! TODO: return error status?  
      FD4_L3_ASSERT(nbext(0)/=bext(0),'neigh(2,1) nbext(0)/=bext(0)')
      FD4_L3_ASSERT(nbext(1)/=bext(1),'neigh(2,1) nbext(1)/=bext(1)')
      FD4_L3_ASSERT(nbext(3)/=bext(3),'neigh(2,1) nbext(3)/=bext(3)')      
      array( 1:bext(0) , 1+ng(1):bext(1)+ng(1) , 1:ng(2) , 1+ng(3):bext(3)+ng(3) ) = & 
           iter%cur%neigh(2,1)%l%fields(idx,st)%l( : , : , nbext(2)-ng(2)+1:nbext(2) , : )
    else
      array( 1:bext(0) , 1+ng(1):bext(1)+ng(1) , 1:ng(2) , 1+ng(3):bext(3)+ng(3) ) = & 
           iter%domain%vartab(idx)%vnull
    end if

    ! ( 0, 1, 0)
    if(associated(iter%cur%neigh(2,2)%l)) then
      nbext = ubound( iter%cur%neigh(2,2)%l%fields(idx,st)%l )
      ! TODO: return error status?  
      FD4_L3_ASSERT(nbext(0)/=bext(0),'neigh(2,2) nbext(0)/=bext(0)')
      FD4_L3_ASSERT(nbext(1)/=bext(1),'neigh(2,2) nbext(1)/=bext(1)')
      FD4_L3_ASSERT(nbext(3)/=bext(3),'neigh(2,2) nbext(3)/=bext(3)')      
      array( 1:bext(0) , 1+ng(1):bext(1)+ng(1) , bext(2)+ng(2)+1:bext(2)+2*ng(2) , 1+ng(3):bext(3)+ng(3) ) = & 
           iter%cur%neigh(2,2)%l%fields(idx,st)%l( : , : , 1:ng(2) , : )
    else
      array( 1:bext(0) , 1+ng(1):bext(1)+ng(1) , bext(2)+ng(2)+1:bext(2)+2*ng(2) , 1+ng(3):bext(3)+ng(3) ) = & 
           iter%domain%vartab(idx)%vnull
    end if

    ! ( 0, 0,-1)
    if(associated(iter%cur%neigh(3,1)%l)) then
      nbext = ubound( iter%cur%neigh(3,1)%l%fields(idx,st)%l )
      ! TODO: return error status?  
      FD4_L3_ASSERT(nbext(0)/=bext(0),'neigh(3,1) nbext(0)/=bext(0)')
      FD4_L3_ASSERT(nbext(1)/=bext(1),'neigh(3,1) nbext(1)/=bext(1)')
      FD4_L3_ASSERT(nbext(2)/=bext(2),'neigh(3,1) nbext(2)/=bext(2)')
      array( 1:bext(0) , 1+ng(1):bext(1)+ng(1) , 1+ng(2):bext(2)+ng(2) , 1:ng(3)) = & 
           iter%cur%neigh(3,1)%l%fields(idx,st)%l( : , : , : , nbext(3)-ng(3)+1:nbext(3) )
    else
      array( 1:bext(0) , 1+ng(1):bext(1)+ng(1) , 1+ng(2):bext(2)+ng(2) , 1:ng(3)) = & 
           iter%domain%vartab(idx)%vnull
    end if

    ! ( 0, 0, 1)
    if(associated(iter%cur%neigh(3,2)%l)) then
      nbext = ubound( iter%cur%neigh(3,2)%l%fields(idx,st)%l )
      ! TODO: return error status?  
      FD4_L3_ASSERT(nbext(0)/=bext(0),'neigh(3,2) nbext(0)/=bext(0)')
      FD4_L3_ASSERT(nbext(1)/=bext(1),'neigh(3,2) nbext(1)/=bext(1)')
      FD4_L3_ASSERT(nbext(2)/=bext(2),'neigh(3,2) nbext(2)/=bext(2)')
      array( 1:bext(0) , 1+ng(1):bext(1)+ng(1) , 1+ng(2):bext(2)+ng(2) , bext(3)+ng(3)+1:bext(3)+2*ng(3) ) = & 
           iter%cur%neigh(3,2)%l%fields(idx,st)%l( : , : , : , 1:ng(3) )
    else
      array( 1:bext(0) , 1+ng(1):bext(1)+ng(1) , 1+ng(2):bext(2)+ng(2) , bext(3)+ng(3)+1:bext(3)+2*ng(3) ) = & 
           iter%domain%vartab(idx)%vnull
    end if


    if(present(opt_bext)) then
      opt_bext(0)   = bext(0)
      opt_bext(1:3) = bext(1:3) + 2 * ng(1:3)
    end if
    
  end subroutine fd4_iter_get_ghost








  !! Get data from the current block with boundary, multiply variables version.
  !!
  !! This is a special version of $fd4_iter_get_ghost.
  !! Get data from variables $idx0 to $idx1 (according to variable table) on the same time step $st.
  !!
  !! *Caution*:
  !! Currently, $vnull is taken from variable $idx0 for *all* variables.
  !!
  subroutine fd4_iter_get_ghost_multi(iter, idx0, idx1, st, aext, array, opt_bext)
    implicit none  
    type(fd4_iter), intent(in)    :: iter       !! the iterator
    integer(i_k),   intent(in)    :: idx0       !! first field index in vartab
    integer(i_k),   intent(in)    :: idx1       !! last field index in vartab
    integer(i_k),   intent(in)    :: st         !! time step index
    integer(i_k),   intent(in)    :: aext(0:3)  !! extents of the buffer array
    real(r_k),      intent(inout) :: array(aext(0),aext(1),aext(2),aext(3))    !! buffer array to write data
    integer(i_k),   intent(out)   :: opt_bext(0:3)  !! actual extents of the block written to the buffer array
    optional :: opt_bext
    
    integer(i_k) :: ng(3)               !! = iter%domain%nghosts, number of ghosts cells
    integer(i_k) :: nbext(0:3)          !! upper bounds of a neighbor block array
    integer(i_k) :: bext(0:3)           !! block extent, inclusive bins
    integer(i_k) :: i0, i1
#if ( FD4_VERBOSE_LEVEL >= 3 )
    integer(i_k) :: idx
#endif
    i0 = iter%domain%varinfo%fields(idx0,st,1)  ! start position
    i1 = iter%domain%varinfo%fields(idx1,st,2)  ! end   position
    
    ng = iter%domain%nghosts
    bext(0)   = i1 - i0 + 1
    bext(1:3) = iter%cur%ext(:)
    
#if ( FD4_VERBOSE_LEVEL >= 3 )
      ! error checks
      FD4_L3_ASSERT(.not.associated(iter%cur),'invalid iterator')
      if(idx0<0 .or. idx1>ubound(iter%domain%vartab,1)) then
        FD4_L1_ERROR('field idx not valid')
      end if
      if(idx0>idx1) then
        FD4_L1_ERROR('idx0>idx1')
      end if
      ! TODO: might change in future
      do idx=idx0,idx1
        if(st<0 .or. st>iter%domain%vartab(idx)%nsteps) then
          FD4_L1_ERROR('time step not valid: '//i2s(idx))
        end if
        if(.not.associated(iter%cur%fields(idx,st)%l)) then
          FD4_L1_ERROR('field not allocated: '//i2s(idx))
        end if
        if(iter%domain%vartab(idx)%facevar/=0) then
          FD4_L1_ERROR('no cell-centered variable: '//i2s(idx))
        end if
      end do
      if(bext(0)>aext(0) .or. bext(1)+2*ng(1)>aext(1) .or. bext(2)+2*ng(2)>aext(2) .or. bext(3)+2*ng(3)>aext(3)) then
        FD4_L1_ERROR('array too small')
      end if
#endif


    ! -- copy current block's data --
    array( 1:bext(0) , 1+ng(1):bext(1)+ng(1) , 1+ng(2):bext(2)+ng(2) , 1+ng(3):bext(3)+ng(3) ) = & 
         iter%cur%sdata(0)%l(i0:i1,:,:,:)

    
    ! -- copy neighbor blocks' data --

    ! TODO: if this field is dynamic, it might be that the neighbor block does not have allocated it!
    
    ! (-1, 0, 0)
    if(associated(iter%cur%neigh(1,1)%l)) then
      nbext = ubound( iter%cur%neigh(1,1)%l%fields(idx0,st)%l )
      ! TODO: return error status?  
      FD4_L3_ASSERT(nbext(2)/=bext(2),'neigh(1,1) nbext(2)/=bext(2)')
      FD4_L3_ASSERT(nbext(3)/=bext(3),'neigh(1,1) nbext(3)/=bext(3)')      
      array( 1:bext(0) , 1:ng(1) , 1+ng(2):bext(2)+ng(2) , 1+ng(3):bext(3)+ng(3) ) = & 
           iter%cur%neigh(1,1)%l%sdata(0)%l( i0:i1 , nbext(1)-ng(1)+1:nbext(1) , : , :)
    else
      array( 1:bext(0) , 1:ng(1) , 1+ng(2):bext(2)+ng(2) , 1+ng(3):bext(3)+ng(3) ) = &
           iter%domain%vartab(idx0)%vnull
    end if

    ! ( 1, 0, 0)
    if(associated(iter%cur%neigh(1,2)%l)) then
      nbext = ubound( iter%cur%neigh(1,2)%l%fields(idx0,st)%l )
      ! TODO: return error status?  
      FD4_L3_ASSERT(nbext(2)/=bext(2),'neigh(1,2) nbext(2)/=bext(2)')      
      FD4_L3_ASSERT(nbext(3)/=bext(3),'neigh(1,2) nbext(3)/=bext(3)')
      array( 1:bext(0) , bext(1)+ng(1)+1:bext(1)+2*ng(1) , 1+ng(2):bext(2)+ng(2) , 1+ng(3):bext(3)+ng(3) ) = & 
           iter%cur%neigh(1,2)%l%sdata(0)%l( i0:i1 , 1:ng(1) , : , : )
    else
      array( 1:bext(0) , bext(1)+ng(1)+1:bext(1)+2*ng(1) , 1+ng(2):bext(2)+ng(2) , 1+ng(3):bext(3)+ng(3) ) = & 
           iter%domain%vartab(idx0)%vnull
    end if
    
    ! ( 0,-1, 0)
    if(associated(iter%cur%neigh(2,1)%l)) then
      nbext = ubound( iter%cur%neigh(2,1)%l%fields(idx0,st)%l )
      ! TODO: return error status?  
      FD4_L3_ASSERT(nbext(1)/=bext(1),'neigh(2,1) nbext(1)/=bext(1)')
      FD4_L3_ASSERT(nbext(3)/=bext(3),'neigh(2,1) nbext(3)/=bext(3)')      
      array( 1:bext(0) , 1+ng(1):bext(1)+ng(1) , 1:ng(2) , 1+ng(3):bext(3)+ng(3) ) = & 
           iter%cur%neigh(2,1)%l%sdata(0)%l( i0:i1 , : , nbext(2)-ng(2)+1:nbext(2) , : )
    else
      array( 1:bext(0) , 1+ng(1):bext(1)+ng(1) , 1:ng(2) , 1+ng(3):bext(3)+ng(3) ) = & 
           iter%domain%vartab(idx0)%vnull
    end if

    ! ( 0, 1, 0)
    if(associated(iter%cur%neigh(2,2)%l)) then
      nbext = ubound( iter%cur%neigh(2,2)%l%fields(idx0,st)%l )
      ! TODO: return error status?  
      FD4_L3_ASSERT(nbext(1)/=bext(1),'neigh(2,2) nbext(1)/=bext(1)')
      FD4_L3_ASSERT(nbext(3)/=bext(3),'neigh(2,2) nbext(3)/=bext(3)')      
      array( 1:bext(0) , 1+ng(1):bext(1)+ng(1) , bext(2)+ng(2)+1:bext(2)+2*ng(2) , 1+ng(3):bext(3)+ng(3) ) = & 
           iter%cur%neigh(2,2)%l%sdata(0)%l( i0:i1 , : , 1:ng(2) , : )
    else
      array( 1:bext(0) , 1+ng(1):bext(1)+ng(1) , bext(2)+ng(2)+1:bext(2)+2*ng(2) , 1+ng(3):bext(3)+ng(3) ) = & 
           iter%domain%vartab(idx0)%vnull
    end if

    ! ( 0, 0,-1)
    if(associated(iter%cur%neigh(3,1)%l)) then
      nbext = ubound( iter%cur%neigh(3,1)%l%fields(idx0,st)%l )
      ! TODO: return error status?  
      FD4_L3_ASSERT(nbext(1)/=bext(1),'neigh(3,1) nbext(1)/=bext(1)')
      FD4_L3_ASSERT(nbext(2)/=bext(2),'neigh(3,1) nbext(2)/=bext(2)')
      array( 1:bext(0) , 1+ng(1):bext(1)+ng(1) , 1+ng(2):bext(2)+ng(2) , 1:ng(3)) = & 
           iter%cur%neigh(3,1)%l%sdata(0)%l( i0:i1 , : , : , nbext(3)-ng(3)+1:nbext(3) )
    else
      array( 1:bext(0) , 1+ng(1):bext(1)+ng(1) , 1+ng(2):bext(2)+ng(2) , 1:ng(3)) = & 
           iter%domain%vartab(idx0)%vnull
    end if

    ! ( 0, 0, 1)
    if(associated(iter%cur%neigh(3,2)%l)) then
      nbext = ubound( iter%cur%neigh(3,2)%l%fields(idx0,st)%l )
      ! TODO: return error status?  
      FD4_L3_ASSERT(nbext(1)/=bext(1),'neigh(3,2) nbext(1)/=bext(1)')
      FD4_L3_ASSERT(nbext(2)/=bext(2),'neigh(3,2) nbext(2)/=bext(2)')
      array( 1:bext(0) , 1+ng(1):bext(1)+ng(1) , 1+ng(2):bext(2)+ng(2) , bext(3)+ng(3)+1:bext(3)+2*ng(3) ) = & 
           iter%cur%neigh(3,2)%l%sdata(0)%l( i0:i1 , : , : , 1:ng(3) )
    else
      array( 1:bext(0) , 1+ng(1):bext(1)+ng(1) , 1+ng(2):bext(2)+ng(2) , bext(3)+ng(3)+1:bext(3)+2*ng(3) ) = & 
           iter%domain%vartab(idx0)%vnull
    end if


    if(present(opt_bext)) then
      opt_bext(0)   = bext(0)
      opt_bext(1:3) = bext(1:3) + 2 * ng(1:3)
    end if
    
  end subroutine fd4_iter_get_ghost_multi






  !! Get cell-face data from the current block.
  !!
  !! Currently, this routine does exactly the same as fd4_iter_get.
  !! This is the order of the indices in $aext, $bext, and $array: (bins, x, y, z)
  !!
  !! *CAUTION*: Error checking will be enabled only if verbose level >= 3.
  !! If not: *Absolutely no checking for error conditions will be carried through*!
  !! This routine will intentionally cause a segmentation violation if the iterator
  !! is not valid, the field $idx does not exist, or the array is to small!
  !!
  !! *TODO*:
  !!  - do we need $fd4_iter_put_facevar ?
  !!  - see $fd4_iter_get
  subroutine fd4_iter_get_facevar(iter, idx, st, aext, array, opt_bext)
    implicit none  
    type(fd4_iter), intent(in)    :: iter       !! the iterator
    integer(i_k),   intent(in)    :: idx        !! field index in vartab
    integer(i_k),   intent(in)    :: st         !! time step index
    integer(i_k),   intent(in)    :: aext(0:3)  !! extents of the buffer array
    real(r_k),      intent(inout) :: array(aext(0),aext(1),aext(2),aext(3))    !! buffer array to write data
    integer(i_k),   intent(out)   :: opt_bext(0:3)  !! actual extents of the block written to the buffer array
    optional :: opt_bext
    
    integer(i_k) :: bext(0:3)        !! block extent, inclusive bins

    !bext = ubound(iter%cur%fields(idx,st)%l)
    bext(0)   = iter%domain%vartab(idx)%nbins
    bext(1:3) = iter%cur%ext(:)
    bext(iter%domain%vartab(idx)%facevar) = bext(iter%domain%vartab(idx)%facevar) + 1
    
#if ( FD4_VERBOSE_LEVEL >= 3 )
      ! error checks
      FD4_L3_ASSERT(.not.associated(iter%cur),'invalid iterator')
      if(idx<0 .or. idx>ubound(iter%domain%vartab,1)) then
        FD4_L1_ERROR('field idx not valid')
      end if
      if(st<0 .or. st>iter%domain%vartab(idx)%nsteps) then
        FD4_L1_ERROR('time step not valid')
      end if
      ! TODO: might change in future
      if(.not.associated(iter%cur%fields(idx,st)%l)) then
        FD4_L1_ERROR('field not allocated')
      end if
      ! check if this var is really a face var
      if(iter%domain%vartab(idx)%facevar<1 .or. iter%domain%vartab(idx)%facevar>3) then
        FD4_L1_ERROR('not a face variable')
      end if
      if(bext(0)>aext(0) .or. bext(1)>aext(1) .or. bext(2)>aext(2) .or. bext(3)>aext(3)) then
        FD4_L1_ERROR('array too small')
      end if
#endif

    array(1:bext(0),1:bext(1),1:bext(2),1:bext(3)) = iter%cur%fields(idx,st)%l
    
    if(present(opt_bext)) opt_bext = bext
        
  end subroutine fd4_iter_get_facevar



  !! Put data with boundary to the current block.
  !!
  !! This is the order of the indices in $aext, $bext, and $array: (bins, x, y, z)
  !!
  !! *CAUTION*: Error checking will be enabled only if verbose level >= 3.
  !! If not: *Absolutely no checking for error conditions will be carried through*!
  !! This routine will intentionally cause a segmentation violation if the iterator
  !! is not valid, the field $idx does not exist, or the array is to small!
  !!
  !! *TODO*:
  !!  - see $fd4_iter_get_ghost
  !!
  subroutine fd4_iter_put_ghost(iter, idx, st, aext, array, dyn)
    implicit none  
    type(fd4_iter),    intent(inout) :: iter         !! the iterator
    integer(i_k),      intent(in)    :: idx          !! field index in vartab
    integer(i_k),      intent(in)    :: st           !! time step index
    integer(i_k),      intent(in)    :: aext(0:3)    !! extents of the buffer array
    real(r_k),         intent(in)    :: array(aext(0),aext(1),aext(2),aext(3))    !! buffer array to read data
    logical, optional, intent(in)    :: dyn          !! do costly checks for dynamic block creation and removal?
    
    integer(i_k) :: bext(0:3)    !! actual extents of the block written to the buffer array
    integer(i_k) :: ng(3)        !! =iter%domain%nghosts, number of ghosts cells
    
    ng = iter%domain%nghosts
    bext(0)   = iter%domain%vartab(idx)%nbins
    bext(1:3) = iter%cur%ext(:)

#if ( FD4_VERBOSE_LEVEL >= 3 )
      ! error checks
      FD4_L3_ASSERT(.not.associated(iter%cur),'invalid iterator')
      if(idx<0 .or. idx>ubound(iter%domain%vartab,1)) then
        FD4_L1_ERROR('field idx not valid')
      end if
      if(st<0 .or. st>iter%domain%vartab(idx)%nsteps) then
        FD4_L1_ERROR('time step not valid')
      end if
      ! TODO: might change in future
      if(.not.associated(iter%cur%fields(idx,st)%l)) then
        FD4_L1_ERROR('field not allocated')
      end if
      if(iter%domain%vartab(idx)%facevar/=0) then
        FD4_L1_ERROR('no cell-centered variable')
      end if
      if(bext(0)>aext(0) .or. bext(1)+2*ng(1)>aext(1) .or. bext(2)+2*ng(2)>aext(2) .or. bext(3)+2*ng(3)>aext(3)) then
        FD4_L1_ERROR('array too small')
      end if
#endif

    ! -- copy current block's data --
    iter%cur%fields(idx,st)%l(:,:,:,:) = &
        array( 1:bext(0), 1+ng(1):bext(1)+ng(1), 1+ng(2):bext(2)+ng(2), 1+ng(3):bext(3)+ng(3) )

    if(present(dyn)) then
      ! check if block is empty 
      if(dyn .and. iter%domain%vartab(idx)%vthres/=FD4_NOTHRES) &
        call fd4_block_empty(iter%cur, iter%domain%vartab, iter%domain%varinfo, idx, st, iter%domain%nghosts)
    end if
        
  end subroutine fd4_iter_put_ghost


  !! Perform empty check on specified variable.
  !! Call this routine if you write values directly in the block data structure
  !! without using $fd4_iter_put_ghost.
  !! The variable must be one with a threshold (vartab%vthres).
  !!
  subroutine fd4_iter_empty(iter, idx, st)
    implicit none  
    type(fd4_iter),    intent(inout) :: iter         !! the iterator
    integer(i_k),      intent(in)    :: idx          !! field index in vartab
    integer(i_k),      intent(in)    :: st           !! time step index

#if ( FD4_VERBOSE_LEVEL >= 3 )
      ! error checks
      FD4_L3_ASSERT(.not.associated(iter%cur),'invalid iterator')
      if(idx<0 .or. idx>ubound(iter%domain%vartab,1)) then
        FD4_L1_ERROR('field idx not valid')
      end if
      if(st<0 .or. st>iter%domain%vartab(idx)%nsteps) then
        FD4_L1_ERROR('time step not valid')
      end if
      if(iter%domain%vartab(idx)%facevar/=0) then
        FD4_L1_ERROR('no cell-centered variable')
      end if
      if(iter%domain%vartab(idx)%vthres==FD4_NOTHRES) then
        FD4_L1_ERROR('vthres == FD4_NOTHRES')
      end if
#endif

    ! check if block is empty 
    call fd4_block_empty(iter%cur, iter%domain%vartab, iter%domain%varinfo, idx, st, iter%domain%nghosts)
        
  end subroutine fd4_iter_empty


end module fd4_iter_mod
