! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010-2012 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!                        FD4 MPI context                       !
!                                                              !
!--------------------------------------------------------------!

#include "fd4.h"

!! FD4 MPI context.
!!
!! - defines MPI data types according to kinds_mod
!! - communicator, rank, nprocs
!! - helper variables, handles, requests, etc. for MPI routines
!! - other persistent fields related to communication
!!
!! *TODO*:
!! - check if we still need all this, see also $fd4_domain because of (de)allocation
!! - good MPIs provide MPI_SIZEOF and MPI_TYPE_MATCH_SIZE to built types matching
!!   all Fortran90 types
!!
!! @see kinds_mod
!! @author Matthias Lieber
module fd4_mpi_mod

  use fd4_globaldef_mod
  use stack_mod
  
  USE_MPI
  
  implicit none
  private
  
  INCLUDE_MPIFH
  
  
  public :: fd4_mpi, stack, fd4_mpi_realloc_array
  public :: MPI_I4K, MPI_I8K, MPI_I_K, MPI_R4K, MPI_R8K, MPI_R_K, MPI_R_K_size
  

  ! MPI data types according to kinds_mod
  
  integer, parameter :: MPI_I4K = MPI_INTEGER    !! 4 byte integer MPI datatype
  integer, parameter :: MPI_I8K = MPI_INTEGER8   !! 8 byte integer MPI datatyte
  integer, parameter :: MPI_I_K = MPI_INTEGER    !! default integer MPI datatyte
  integer, parameter :: MPI_R4K = MPI_REAL              !! 4 byte real MPI datatype
  integer, parameter :: MPI_R8K = MPI_DOUBLE_PRECISION  !! 8 byte real MPI datatyte
  integer, parameter :: MPI_R_K = MPI_DOUBLE_PRECISION  !! default real MPI datatyte

  integer(i_k), parameter :: MPI_R_K_size = 8 !! size in bytes of a variable of type MPI_R_K
  

  type fd4_mpi
    integer :: comm       !! MPI communicator
    integer :: comm_gc    !! MPI communicator for ghost communication
    integer :: comm_cpl   !! MPI communicator for coupling
    integer :: comm_mig   !! MPI communicator for block migration
    integer :: rank = -1  !! MPI rank
    integer :: nprocs     !! number of MPI processes
    integer :: lognprocs  !! ceiling(log(nprocs))
    
    integer, pointer :: recvcnt(:) => NULL() !! for MPI_Allgatherv
    integer, pointer :: recvdis(:) => NULL() !! for MPI_Allgatherv
    integer, pointer :: senddis(:) => NULL() !! for MPI_Allgatherv
    
    real(r4k), pointer :: sendreal(:) => NULL() !! to exchange integers with neighbors
    real(r4k), pointer :: recvreal(:) => NULL() !! to exchange integers with neighbors

    type(stack) :: requests               !! requests for non-blocking communication

    integer, pointer :: statuses(:,:)     !! array of statuses
    
  end type fd4_mpi
  
contains

  !! Reallocate the $array if necessary (i.e. $size($array,1) < $sz or not allocated).
  !! Any array content gets lost.
  !!
  subroutine fd4_mpi_realloc_array(array, sz, ubound, err)
    integer, pointer :: array(:)
    integer(i_k), intent(in) :: sz, ubound
    integer(i_k), intent(out) :: err
    err = -1
    if(associated(array)) then
      if(size(array,1)<sz) then
        deallocate(array)
        allocate(array(ubound:sz-1+ubound), stat=err )
        FD4_L1_ASSERT(err/=0,'allocation failed, sz='//i2s(sz))
        !FD4_L2_INFO('realloc, sz='//i2s(sz))
      end if
    end if
    if(.not. associated(array)) then
      allocate(array(ubound:sz-1+ubound), stat=err )
      FD4_L1_ASSERT(err/=0,'allocation failed, sz='//i2s(sz))
      !FD4_L2_INFO('alloc, sz='//i2s(sz))
    end if
    err = 0
  end subroutine fd4_mpi_realloc_array
  
end module fd4_mpi_mod
