! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010-2012 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!                     NetCDF4 Parallel I/O                     !
!                                                              !
!--------------------------------------------------------------!

#include "fd4.h"

!! Write data of an $fd4_domain to a NetCDF4/HDF5 file.
!!
!! How it works:
!! Create couple context to map the data into a new decomposition which is better suited for I/O.
!! Ideally only decomposed in last dimension (z) with one block for each I/O rank.
!! This may result in much less blocks than ranks, which could require too much memory per I/O rank.
!! In these cases, decomposition in more than one dimensions in required.
!!
!! *TODO*:
!!  - open/close file for each $fd4_netcdf4_write call (can be opened before normal program termination)
!!  - how to visualize 5D data? (bin, x, y, z, time)
!!  - support for multiple bin dimensions (i.e. vars with different number of bins)
!!    - currently, one bin dimension with max size is used
!!  - optional attributes (dimension names, grid cell sizes, lat/lon/level, etc.)
!!  - better handling of face variables?
!!    - currently, interpolation to center or neglecting additional row is possible
!!  - let user choose between 'single file' mode and 'file for each output step' mode
!!  - more error/arguments checking
!!
!! @author Matthias Lieber
module fd4_netcdf4_mod

  use fd4_globaldef_mod
  use fd4_vartab_mod
  use fd4_domain_mod
  use fd4_metadata_mod
  use fd4_couple_mod
  use fd4_mpi_mod
  use netcdf
  
  USE_MPI
  
  implicit none
  private
  
  INCLUDE_MPIFH
   
  ! public methods
  public :: fd4_netcdf4_open
  public :: fd4_netcdf4_write
  public :: fd4_netcdf4_close
  
  public :: fd4_netcdf4_comm, fd4_domain
  public :: fd4_netdf4_step_def, fd4_netcdf4_cplarray
  
  integer(i_k), parameter :: MAX_STEP_DEF = 4
  
  !! type for a coupling array
  type fd4_netcdf4_cplarray
    integer(i_k) :: idx
    real(r_k), pointer :: a3(:,:,:) => null()
    real(r_k), pointer :: a4(:,:,:,:) => null()
  end type fd4_netcdf4_cplarray
  
  !! all data structures depending on the time step indexes of the output variables
  type fd4_netdf4_step_def
    integer(i_k), pointer :: st(:) => null()       !! time step indexes of the data fields to write
    integer(i_k) :: stsum                          !! check sum of the time steps for quick searching
    type(fd4_couple) :: cpl                        !! couple context to map data into decomposotion better suited for I/O
  end type fd4_netdf4_step_def
  
  
  !! FD4 NetCDF4 I/O Communicator
  type fd4_netcdf4_comm
    type(fd4_domain), pointer :: domain => null()  !! the domain which owns the blocks
    integer :: ncid                                !! NetCDF file handle
    integer :: dimids(0:4)                         !! NetCDF dimension IDs
    integer, pointer :: varid(:) => null()         !! NetCDF variable IDs
    integer(i_k) :: rec                            !! current output record (output time step)
    integer(i_k), pointer :: idx(:) => null()      !! field indices of the data fields to write
    logical :: interpolate                         !! interpolate face vars to center?
    integer(i_k) :: maxnbins                       !! max number of bins
    !
    type(fd4_netdf4_step_def) :: sd(MAX_STEP_DEF)  !! couple context and time step indexes for each step definition
    integer(i_k) :: decomp(3)                      !! rectangular 3D domain decomposition for I/O
    integer :: iocomm = MPI_COMM_NULL              !! MPI communicator for processes performing I/O
    integer :: nioprocs = 0                        !! number of I/O processes
    integer :: iorank = -1                         !! rank in iocomm
    type(fd4_netcdf4_cplarray), pointer :: cplarray(:) => null()  !! array of local couple arrays
    integer(i_k) :: abnd(3,2) = -1                 !! bounds of couple arrays of this rank
    integer(i_k) :: ext(3) = -1                    !! extent of couple arrays of this rank
    integer(i_k) :: step_def = 0                   !! number of time step index definitions
    integer, pointer :: io2fd4rank(:) => null()    !! map from io ranks to FD4 ranks (only allocated if req.)
  end type fd4_netcdf4_comm
  

contains


  !! Open a NetCDF4 file.
  !!
  !! - $int_opt is not working currently, only for backward compatiblity.
  !!
  subroutine fd4_netcdf4_open(nfcomm, domain, filename, nfields, idx, st, err, int_opt, opt_decomp, opt_maxvol)
    type(fd4_netcdf4_comm),  intent(inout) :: nfcomm         !! FD4 NetCDF4 I/O Communicator
    type(fd4_domain),        intent(inout) :: domain         !! the domain
    character(*),            intent(in)    :: filename       !! name of the file to write
    integer(i_k),            intent(in)    :: nfields        !! number of data fields to write
    integer(i_k),            intent(in)    :: idx(nfields)   !! field indices of the data fields to write
    integer(i_k),            intent(in)    :: st(nfields)    !! time step indices of the data fields to write
    integer(i_k),            intent(out)   :: err            !! error status: 0...ok    
    logical,       optional, intent(in)    :: int_opt        !! interpolate face vars to center?
    integer(i_k),  optional, intent(in)    :: opt_decomp(3)  !! I/O decomposition (default: full in z direction)
    integer(i8k),  optional, intent(in)    :: opt_maxvol     !! maximum volume of one I/O partition in byte (default: 128MiB)
    target :: nfcomm, domain
    
    integer(i_k) :: dext(0:3), i, color, dim
    integer(i8k) :: cellvol, vol, maxvol
    
    ! NetCDF4 variables
    integer :: stat
    character(FD4_VARNAME_LEN+8) :: varname
        
    err = -1
    
    FD4_L1_ASSERT(associated(nfcomm%domain),'NetCDF4 communicator is in use')
    nfcomm%domain => domain
    
    nfcomm%interpolate = .false.
    if(present(int_opt)) nfcomm%interpolate = int_opt
        
    ! spatial domain size
    dext(1:3) = domain%dbnd(1:3,2) - domain%dbnd(1:3,1) + 1

    ! copy idx and st (variable index and step)
    allocate(nfcomm%idx(nfields))
    nfcomm%idx = idx
    
    if(present(opt_decomp)) then
      nfcomm%decomp(1:3) = opt_decomp(1:3)
      do i=1,3
        nfcomm%decomp(i) = MAX(nfcomm%decomp(i), 1)
        nfcomm%decomp(i) = MIN(nfcomm%decomp(i), domain%dbnd(i,2)-domain%dbnd(i,1)+1)
        nfcomm%decomp(i) = MIN(nfcomm%decomp(i), domain%mpi%nprocs)
      end do
      if(PRODUCT(nfcomm%decomp) > domain%mpi%nprocs) then
        ! reduce number of partitions
        dim_loop: do i=1,3
          do while (nfcomm%decomp(i)>1)
            nfcomm%decomp(i) = MAX( nfcomm%decomp(i) / 2, 1)
            if(PRODUCT(nfcomm%decomp) <= domain%mpi%nprocs) exit dim_loop
          end do
        end do dim_loop
      end if
    elseif(ANY(domain%md%mdsdnum>1)) then
      ! metadata subdomains are present
      nfcomm%decomp(1:3) = domain%md%mdsdnum(1:3)
    else
      ! default decomp: as much as possible io partitions in z, but no decomp in x and y
      nfcomm%decomp(1:3) = 1
      nfcomm%decomp(3)   = MIN(dext(3), domain%mpi%nprocs)
    end if

    maxvol = 128*2**20 ! 128 MiB
    if(present(opt_maxvol)) maxvol=opt_maxvol
    ! determine volume per grid cell
    cellvol = 0_i8k
    do i = 1, nfields
      cellvol = cellvol + domain%vartab(idx(i))%nbins
    end do
    cellvol = cellvol * MPI_R_K_size     ! size of grid variables
    cellvol = cellvol + MPI_R_K_size + 4 ! add size of block weight and block map
    ! estimate memory usage for couple arrays, if necessary increase number of io partitions
    dim=3 ! first dimension to further decompose
    do while(dim>0)
      vol = cellvol * PRODUCT(dext(1:3) / nfcomm%decomp(1:3)) ! multiply with approx. io partition size
      if(vol<=maxvol .or. PRODUCT(nfcomm%decomp)*2 > domain%mpi%nprocs) exit
      if(dext(dim)>=2*nfcomm%decomp(dim)) then
        nfcomm%decomp(dim)=2*nfcomm%decomp(dim)
      else
        dim=dim-1
      end if
    end do
    
    if(domain%mpi%rank==0) FD4_L4_INFO('netcdf4 output decomp: '//p2s(nfcomm%decomp))
    
    nfcomm%nioprocs = PRODUCT(nfcomm%decomp)

    if(nfcomm%nioprocs > domain%mpi%nprocs) then
      FD4_L1_ERROR('wrong decomp: '//p2s(nfcomm%decomp))
    end if
    
    ! build I/O communicator
    if(nfcomm%nioprocs < domain%mpi%nprocs) then
      ! MPI_Comm_split if PRODUCT(decomp) < domain%mpi%nprocs
      color = MPI_UNDEFINED
      allocate(nfcomm%io2fd4rank(0:nfcomm%nioprocs-1))
      do i=0,nfcomm%nioprocs-1
        nfcomm%io2fd4rank(i) = (domain%mpi%nprocs*i)/nfcomm%nioprocs
        if(nfcomm%io2fd4rank(i)==domain%mpi%rank) color = 1
      end do
      FD4_L3_ASSERT(MINVAL(nfcomm%io2fd4rank)<0,'io2fd4rank invalid')
      FD4_L3_ASSERT(MAXVAL(nfcomm%io2fd4rank)>=domain%mpi%nprocs,'io2fd4rank invalid')
      call MPI_Comm_split(domain%mpi%comm, color, domain%mpi%rank, nfcomm%iocomm, err)
      err = err - MPI_SUCCESS
      FD4_L1_ASSERT(err/=0,'MPI_Comm_split failed')
      if(color == MPI_UNDEFINED) then
        nfcomm%iorank = -1
      else
        call MPI_Comm_rank(nfcomm%iocomm,nfcomm%iorank,err) 
        call MPI_Comm_size(nfcomm%iocomm,i,err)
        FD4_L1_ASSERT(nfcomm%nioprocs/=i,'invald iocomm size')
      end if
    else
      call MPI_Comm_dup(domain%mpi%comm,nfcomm%iocomm,err)
      err = err - MPI_SUCCESS
      FD4_L1_ASSERT(err/=0,'MPI_Comm_dup failed')
      call MPI_Comm_rank(nfcomm%iocomm,nfcomm%iorank,err) 
      call MPI_Comm_size(nfcomm%iocomm,i,err)
      FD4_L1_ASSERT(nfcomm%nioprocs/=i,'invald iocomm size')
    end if

    nfcomm%abnd(1:3,1) =  0
    nfcomm%abnd(1:3,2) = -1
    nfcomm%step_def = 0
    
    if(nfcomm%iorank >= 0) allocate(nfcomm%cplarray(nfields))
    
    ! creates couple context for given time step indexes
    call fd4_netcdf4_add_step_def(nfcomm, st, err)
    FD4_L1_ASSERT(err/=0,'adding step def failed')
    
    ! I/O ranks only
    if(nfcomm%iorank >= 0) then

      ! allocate NetCDF variable id
      allocate(nfcomm%varid(nfields+2))

      ! Create the netCDF file. 
      stat = nf90_create_par( filename, IOR( NF90_HDF5, NF90_CLASSIC_MODEL ) , & 
                              nfcomm%iocomm, MPI_INFO_NULL, nfcomm%ncid )
      err = stat - NF90_NOERR
      FD4_L1_ASSERT(err/=0,'nf90_create: '//trim(nf90_strerror(stat)))

      ! get max number of bins
      dext(0) = 1
      do i=1,size(domain%vartab)
        dext(0) = MAX(domain%vartab(i)%nbins,dext(0))
      end do
      nfcomm%maxnbins = dext(0)

      !! define the dimensions !!

      ! bin dimension
      stat = nf90_def_dim(nfcomm%ncid, 'bin', dext(0), nfcomm%dimids(0))
      err = stat - NF90_NOERR
      FD4_L1_ASSERT(err/=0,'nf90_def_dim: '//trim(nf90_strerror(stat)))

      ! 3 spatial dimensions
      stat = nf90_def_dim(nfcomm%ncid, 'x', dext(1), nfcomm%dimids(1))
      err = stat - NF90_NOERR
      FD4_L1_ASSERT(err/=0,'nf90_def_dim: '//trim(nf90_strerror(stat)))

      stat = nf90_def_dim(nfcomm%ncid, 'y', dext(2), nfcomm%dimids(2))
      err = stat - NF90_NOERR
      FD4_L1_ASSERT(err/=0,'nf90_def_dim: '//trim(nf90_strerror(stat)))

      stat = nf90_def_dim(nfcomm%ncid, 'z', dext(3), nfcomm%dimids(3))
      err = stat - NF90_NOERR
      FD4_L1_ASSERT(err/=0,'nf90_def_dim: '//trim(nf90_strerror(stat)))

      ! unlimited time dimension (records)
      stat = nf90_def_dim(nfcomm%ncid, 'time', NF90_UNLIMITED, nfcomm%dimids(4))
      err = stat - NF90_NOERR
      FD4_L1_ASSERT(err/=0,'nf90_def_dim: '//trim(nf90_strerror(stat)))


      ! define variables and fill values
      do i = 1, nfields

        varname = domain%vartab(idx(i))%name
        if(COUNT(idx == idx(i))>1) then
          ! variable idx appears multiple times in idx - append step
          varname = STRIM(varname)//'_'//i2s(st(i))
        end if
        ! define the variable, set fill value, set independent access mode
        call fd4_netcdf4_def_var(varname, domain%vartab(idx(i))%vnull, &
                                 domain%vartab(idx(i))%nbins > 1, nfcomm%varid(i), err)
        if(err/=0) return

      end do

      ! additional variable for the block partitioning
      call fd4_netcdf4_def_var_int('blocks', -1, nfcomm%varid(nfields+1), err)
      if(err/=0) return

      ! additional variable for the block partitioning
      call fd4_netcdf4_def_var('weight', 0.0_r_k, .false., nfcomm%varid(nfields+2), err)
      if(err/=0) return

      ! End define mode. This tells netCDF we are done defining metadata.
      stat = nf90_enddef(nfcomm%ncid)
      err = stat - NF90_NOERR
      FD4_L1_ASSERT(err/=0,'nf90_enddef: '//trim(nf90_strerror(stat)))
    
    end if
    
    ! set current record to 0
    nfcomm%rec = 0
    
    if(FD4_VERBOSE_LEVEL >= 2 .and. domain%mpi%rank == 0) then
      write (fd4_out,'(A,A,A,A,3(I4),A)') FD4_LINEHD,'opened file ',STRIM(filename), &
                                          ' with ',nfcomm%decomp(1:3),' processes'
    end if    
    
    err = 0

  contains

    !! define a real(r_k) variable with given name and vnull
    !!
    !! internal subroutine
    !!
    subroutine fd4_netcdf4_def_var(name, vnull, hasbins, varid, err)
      character(*), intent(in) :: name
      real(r_k), intent(in)    :: vnull
      logical, intent(in)      :: hasbins
      integer, intent(out)     :: varid
      integer, intent(inout)   :: err
      
      integer :: nf90_datatype, ld
      
      if(r_k == r8k) then
        nf90_datatype = NF90_DOUBLE
      elseif(r_k == r4k) then
        nf90_datatype = NF90_FLOAT
      else
        err = -1
        FD4_L1_ERROR('unknown real kind r_k')
      end if
      
      ! determine lowest dimension: 0=bins, 1=x, 2=y, 3=z, 4=time
      ld = 1    
      if(hasbins) ld = 0
    
      ! define the variable
      stat = nf90_def_var(nfcomm%ncid, name, nf90_datatype, nfcomm%dimids(ld:4), varid)
      err = stat - NF90_NOERR
      FD4_L1_ASSERT(err/=0,'nf90_def_var: '//trim(nf90_strerror(stat)))
      
      ! define the fill value for the variable
      stat = nf90_put_att(nfcomm%ncid, varid, '_FillValue',  vnull )
      !stat = nf90_def_var_fill(nfcomm%ncid, varid, 0, vnull )
      err = stat - NF90_NOERR
      FD4_L1_ASSERT(err/=0,'nf90_put_att: '//trim(nf90_strerror(stat)))

      ! set independend write mode
      stat = nf90_var_par_access(nfcomm%ncid, varid, NF90_INDEPENDENT)
      err = stat - NF90_NOERR
      FD4_L1_ASSERT(err/=0,'nf90_var_par_access: '//trim(nf90_strerror(stat)))

    end subroutine fd4_netcdf4_def_var


    !! define an integer variable with given name and vnull
    !!
    !! internal subroutine
    !!
    subroutine fd4_netcdf4_def_var_int(name, vnull, varid, err)
      character(*), intent(in) :: name
      integer(i_k), intent(in) :: vnull
      integer, intent(out)     :: varid
      integer, intent(inout)   :: err
    
      ! define the variable
      stat = nf90_def_var(nfcomm%ncid, name, NF90_INT, nfcomm%dimids(1:4), varid)
      err = stat - NF90_NOERR
      FD4_L1_ASSERT(err/=0,'nf90_def_var: '//trim(nf90_strerror(stat)))
      
      ! define the fill value for the variable
      stat = nf90_put_att(nfcomm%ncid, varid, '_FillValue',  vnull )
      !stat = nf90_def_var_fill(nfcomm%ncid, varid, 0, vnull )
      err = stat - NF90_NOERR
      FD4_L1_ASSERT(err/=0,'nf90_put_att: '//trim(nf90_strerror(stat)))

      ! set independend write mode
      stat = nf90_var_par_access(nfcomm%ncid, varid, NF90_INDEPENDENT)
      err = stat - NF90_NOERR
      FD4_L1_ASSERT(err/=0,'nf90_var_par_access: '//trim(nf90_strerror(stat)))

    end subroutine fd4_netcdf4_def_var_int

    
  end subroutine fd4_netcdf4_open


  !! Add a set of time step indexes for opened file to write.
  !!
  !! *TODO:*
  !!  - re-use couple array bounds from former fd4_couple's of this io communicator.
  !!
  !! Internal routine.
  subroutine fd4_netcdf4_add_step_def(nfcomm, st, err)
    type(fd4_netcdf4_comm), intent(inout), target :: nfcomm    !! FD4 NetCDF4 I/O Communicator
    integer(i_k),           intent(in)    :: st(:)             !! time step indices of the data fields to write
    integer(i_k),           intent(out)   :: err               !! error status: 0...ok    

    integer(i_k) :: abnd(3,2), offset(3), dcpos(3), iorank, dim, i, nfields, fd4rank, cplarrayidx
    type(fd4_couple), pointer :: cpl
    type(fd4_domain), pointer :: domain

    err = -1

    nfields = size(nfcomm%idx)
    
    FD4_L1_ASSERT(nfcomm%nioprocs<=0,'no i/o processes')
    FD4_L1_ASSERT(size(st)<nfields,'too few step indexes')
    if(nfcomm%step_def >= size(nfcomm%sd)) then
      FD4_L1_ERROR('max '//i2s(MAX_STEP_DEF)//' step defs allowed')
    end if
    if(nfcomm%iorank>=0 .and. .not. associated(nfcomm%cplarray)) then
      FD4_L1_ERROR('cplarray not allocated')
    end if
    
    nfcomm%step_def = nfcomm%step_def + 1
    
    domain => nfcomm%domain
    cpl => nfcomm%sd(nfcomm%step_def)%cpl
    call fd4_couple_create(cpl, domain, err, opt_cpldir=FD4_CPL_GET, opt_fext=.false., opt_nabnd=1, opt_nvar=size(nfcomm%idx))
    FD4_L1_ASSERT(err/=0,'fd4_couple_create failed')

    ! allocate step index array and copy argument st
    allocate(nfcomm%sd(nfcomm%step_def)%st(nfields))
    nfcomm%sd(nfcomm%step_def)%st(1:nfields) = st(1:nfields)
    nfcomm%sd(nfcomm%step_def)%stsum = SUM(st(1:nfields))

    ! add couple arrays bounds
    dcpos(1:3) = 0
    do iorank=0,nfcomm%nioprocs-1
      ! transform iorank to fd4rank
      fd4rank = iorank
      if(associated(nfcomm%io2fd4rank)) fd4rank = nfcomm%io2fd4rank(iorank)
      ! determine couple array bounds for this io rank
      do dim=1,3
        if(nfcomm%decomp(dim)==1) then
          abnd(dim,1:2) = domain%dbnd(dim,1:2)
        else
          abnd(dim,1) = domain%dbnd(dim,1) + ( dcpos(dim)   *(domain%dbnd(dim,2)-domain%dbnd(dim,1)+1))/nfcomm%decomp(dim)
          abnd(dim,2) = domain%dbnd(dim,1) + ((dcpos(dim)+1)*(domain%dbnd(dim,2)-domain%dbnd(dim,1)+1))/nfcomm%decomp(dim) - 1
        end if
      end do
      !if(domain%mpi%rank==0) then
      !  fd4rank = iorank
      !  if(associated(nfcomm%io2fd4rank)) fd4rank = nfcomm%io2fd4rank(iorank)
      !  write(*,'(A,I5,A,3(I4),A,3(I4),A,3(I4))') 'rank',fd4rank,'    dcpos',dcpos,'    abnd',abnd(1:3,1),' -',abnd(1:3,2)
      !end if
      if(nfcomm%step_def==1 .and. iorank==nfcomm%iorank) nfcomm%abnd = abnd
      call fd4_couple_add_partition(cpl, fd4rank, abnd, err)
      FD4_L1_ASSERT(err/=0,'fd4_couple_add_partition at fd4rank='//i2s(fd4rank))
      ! next dcpos
      do dim=1,3
        if(dcpos(dim)<nfcomm%decomp(dim)-1) then
          dcpos(dim) = dcpos(dim) + 1
          exit
        end if
        dcpos(dim) = 0
      end do
    end do
    
    ! add variables
    do i=1,size(nfcomm%idx)
      cplarrayidx = -1
      call fd4_couple_add_var(cpl, nfcomm%idx(i), st(i), err, opt_cplarridx = cplarrayidx)
      FD4_L1_ASSERT(err/=0,'fd4_couple_add_var at i='//i2s(i))
      if(nfcomm%iorank >= 0) nfcomm%cplarray(i)%idx = cplarrayidx
    end do
    
    ! I/O ranks only
    if(nfcomm%iorank >= 0) then
      if(nfcomm%step_def==1) then
        nfcomm%ext(1:3) = nfcomm%abnd(1:3,2) - nfcomm%abnd(1:3,1) + 1
        FD4_L1_ASSERT(ANY(nfcomm%ext(1:3)<=0),'nfcomm%abnd not correct')
      end if
      offset(1:3) = 1
      do i=1,size(nfcomm%idx)
        if(domain%vartab(nfcomm%idx(i))%nbins==1) then
          if(nfcomm%step_def==1) then
            allocate(nfcomm%cplarray(i)%a3(nfcomm%ext(1),nfcomm%ext(2),nfcomm%ext(3)))
          end if
          call fd4_couple_set_local_3D_array(cpl, nfcomm%cplarray(i)%idx, nfcomm%cplarray(i)%a3, offset, err)
          FD4_L1_ASSERT(err/=0,'fd4_couple_set_local_3D_array at i='//i2s(i))
        else
          if(nfcomm%step_def==1) then
            allocate(nfcomm%cplarray(i)%a4(domain%vartab(nfcomm%idx(i))%nbins,nfcomm%ext(1),nfcomm%ext(2),nfcomm%ext(3)))
          end if
          call fd4_couple_set_local_4D_array(cpl, nfcomm%cplarray(i)%idx, nfcomm%cplarray(i)%a4, offset, 1, err)
          FD4_L1_ASSERT(err/=0,'fd4_couple_set_local_4D_array at i='//i2s(i))
        end if
      end do
    end if

    call fd4_couple_commit(cpl, err)
    FD4_L1_ASSERT(err/=0,'fd4_couple_commit failed')
    
    !call fd4_couple_debug_metadata(cpl, 'couple-netcdf4-'//i2s(nfcomm%step_def,.true.))
    
    err = 0

  end subroutine fd4_netcdf4_add_step_def

  !! Close NetCDF file.
  subroutine fd4_netcdf4_close(nfcomm, err)
    type(fd4_netcdf4_comm),   intent(inout) :: nfcomm      !! FD4 NetCDF4 I/O Communicator
    integer(i_k),             intent(out)   :: err         !! error status: 0...ok    
    integer :: stat
    
    integer(i_k) :: i
    
    err = -1
    
    ! close NetCDF file - I/O ranks only
    if(nfcomm%iorank >= 0) then
      stat = nf90_close(nfcomm%ncid)
      err = stat - NF90_NOERR
      FD4_L1_ASSERT(err/=0,'nf90_close: '//trim(nf90_strerror(stat)))
    end if

    nfcomm%domain => null()
    
    deallocate(nfcomm%idx)
    if(associated(nfcomm%varid)) deallocate(nfcomm%varid)
    
    ! delete couple contexts
    do i=1,nfcomm%step_def
      if(associated(nfcomm%sd(i)%st)) deallocate(nfcomm%sd(i)%st)
      call fd4_couple_delete(nfcomm%sd(i)%cpl, err)
      FD4_L1_ASSERT(err/=0,'fd4_couple_delete failed i: '//i2s(i))
    end do
    nfcomm%step_def=0
    
    ! free couple arrays
    if(associated(nfcomm%cplarray)) then
      do i=1,size(nfcomm%cplarray)
        if(associated(nfcomm%cplarray(i)%a3)) deallocate(nfcomm%cplarray(i)%a3)
        if(associated(nfcomm%cplarray(i)%a4)) deallocate(nfcomm%cplarray(i)%a4)
      end do
      deallocate(nfcomm%cplarray)
    end if
       
    ! free I/O communicator
    if(nfcomm%iocomm /= MPI_COMM_NULL) then
      call MPI_Comm_free(nfcomm%iocomm,err)
      err = err - MPI_SUCCESS
      FD4_L1_ASSERT(err/=0,'MPI_Comm_free failed')
    end if
    
    nfcomm%iocomm = MPI_COMM_NULL
    nfcomm%iorank = -1
    nfcomm%nioprocs = 0

    err = 0

  end subroutine fd4_netcdf4_close


  !! Write current data of the domain to NetCDF file.
  !!
  !! The optional $st_opt parameter will overwrite the values given
  !! to $fd4_netcdf4_open and create a new set of time step indexes.
  subroutine fd4_netcdf4_write(nfcomm, err, st_opt)
    type(fd4_netcdf4_comm), intent(inout) :: nfcomm    !! FD4 NetCDF4 I/O Communicator
    integer(i_k),           intent(out)   :: err       !! error status: 0...ok    
    integer(i_k), optional, intent(in)    :: st_opt(:) !! time step indices of the data fields to write
    target :: nfcomm, st_opt
    
    integer(i_k) :: i, j, k, nbins, ld, idx, bnd(3,2), ext(3), nfields
    real(r_k), allocatable :: array(:,:,:)
    integer, allocatable :: array_int(:,:,:)
    type(fd4_couple), pointer :: cpl
    
    integer :: start(0:4), count(0:4), stat
    
    err = -1
    
    FD4_L1_ASSERT(nfcomm%step_def==0,'file not opened')
    FD4_L1_ASSERT(nfcomm%nioprocs<=0,'no i/o processes')
    if(nfcomm%iorank>=0 .and. .not. associated(nfcomm%cplarray)) then
      FD4_L1_ERROR('cplarray not allocated')
    end if

    nfields = size(nfcomm%idx)
    
    if(present(st_opt)) then
      cpl => null()
      FD4_L1_ASSERT(size(st_opt)<nfields,'too few step indexes')
      ! search nfcomm%sd for matching step indexes
      do i=1,nfcomm%step_def
        j = SUM(st_opt(1:nfields))
        if( nfcomm%sd(i)%stsum == j ) then
          if( ALL(nfcomm%sd(i)%st(1:nfields) == st_opt(1:nfields)) ) then
            cpl => nfcomm%sd(i)%cpl
            FD4_L5_INFO('found step def '//i2s(i))
            exit
          end if
        end if
      end do
      if(.not. associated(cpl)) then
        ! we need to create a new step definition and couple context
        call fd4_netcdf4_add_step_def(nfcomm, st_opt, err)
        FD4_L1_ASSERT(err/=0,'adding step def failed')
        FD4_L5_INFO('new step def '//i2s(nfcomm%step_def))
        cpl => nfcomm%sd(nfcomm%step_def)%cpl
      end if
    else
      FD4_L5_INFO('default step def 1')
      cpl => nfcomm%sd(1)%cpl
    end if

    !call fd4_couple_debug_metadata(cpl, 'couple-netcdf4-'//i2s(nfcomm%rec,.true.))

    ! transform data into io decomposition using the defined couple context
    call fd4_couple_get (cpl, err)
    FD4_L1_ASSERT(err/=0,'fd4_couple_get failed')
    
    nfcomm%rec = nfcomm%rec + 1
    
    ! only the io ranks write data
    if(nfcomm%iorank >= 0) then

      start(1:3) = nfcomm%abnd(1:3,1) - nfcomm%domain%dbnd(1:3,1) + 1
      count(1:3) = nfcomm%ext(1:3)
      start(4) = nfcomm%rec
      count(4) = 1
    
      do i=1,nfields
        idx = nfcomm%idx(i)
        nbins = nfcomm%domain%vartab(idx)%nbins
        start(0) = 1
        count(0) = nbins
        ld = 1
        if(nbins>1) ld = 0

        if(associated(nfcomm%cplarray(i)%a3)) then
          stat = nf90_put_var(nfcomm%ncid, nfcomm%varid(i), nfcomm%cplarray(i)%a3, start(ld:4), count(ld:4) )
        else if(associated(nfcomm%cplarray(i)%a4)) then
          stat = nf90_put_var(nfcomm%ncid, nfcomm%varid(i), nfcomm%cplarray(i)%a4, start(ld:4), count(ld:4) )
        else
          err = -2
          FD4_L1_ERROR('i/o couple array not allocated')
        end if

        err = stat - NF90_NOERR
        FD4_L1_ASSERT(err/=0,'nf90_put_var: '//trim(nf90_strerror(stat)))      

      end do
      
      ! write block partitioning
      ! and block weights per grid cell
      allocate(array_int(count(1),count(2),count(3)))
      array_int = -10_i_k
      if(associated(nfcomm%domain%balance%blkweights)) then
        allocate(array    (count(1),count(2),count(3)))
        array = -10.0_r_k
      end if


      ! loop over blocks which intersect with couple arrays
      do k=cpl%loopbnd(3,1),cpl%loopbnd(3,2)
        do j=cpl%loopbnd(2,1),cpl%loopbnd(2,2)
          do i=cpl%loopbnd(1,1),cpl%loopbnd(1,2)
            ! create block bounds
            bnd(1,:) = (/ nfcomm%domain%d(1)%bl(i) , nfcomm%domain%d(1)%bl(i+1) - 1 /)
            bnd(2,:) = (/ nfcomm%domain%d(2)%bl(j) , nfcomm%domain%d(2)%bl(j+1) - 1 /)
            bnd(3,:) = (/ nfcomm%domain%d(3)%bl(k) , nfcomm%domain%d(3)%bl(k+1) - 1 /)
            ext(1:3) = bnd(1:3,2) - bnd(1:3,1) + 1
            ! intersect block with couple array
            bnd(1:3,1)=MAX(bnd(1:3,1),nfcomm%abnd(1:3,1))
            bnd(1:3,2)=MIN(bnd(1:3,2),nfcomm%abnd(1:3,2))
            if(bnd(1,1)<=bnd(1,2) .and. &
               bnd(2,1)<=bnd(2,2) .and. &
               bnd(3,1)<=bnd(3,2) ) then
              ! make bnd local to couple array starting with (/1,1,1/)
              bnd(1:3,1) = bnd(1:3,1) - nfcomm%abnd(1:3,1) + 1
              bnd(1:3,2) = bnd(1:3,2) - nfcomm%abnd(1:3,1) + 1
              ! fill blkmap and blkwgt array
              array_int(bnd(1,1):bnd(1,2),bnd(2,1):bnd(2,2),bnd(3,1):bnd(3,2)) = &
                 nfcomm%domain%blkmap(i,j,k,nfcomm%domain%blkmap_now)
              if(associated(nfcomm%domain%balance%blkweights)) then
                array(bnd(1,1):bnd(1,2),bnd(2,1):bnd(2,2),bnd(3,1):bnd(3,2)) = &
                  REAL(nfcomm%domain%balance%blkweights(i,j,k), r_k) / REAL( PRODUCT(ext), r_k)
              end if
            end if
          end do
        end do
      end do
      
      ! write blkmap
      stat = nf90_put_var(nfcomm%ncid, nfcomm%varid(size(nfcomm%varid)-1), array_int, start(1:4), count(1:4) )
      err = stat - NF90_NOERR
      FD4_L1_ASSERT(err/=0,'nf90_put_var: '//trim(nf90_strerror(stat)))
      deallocate(array_int)

      ! write blkwgts
      if(associated(nfcomm%domain%balance%blkweights)) then
        stat = nf90_put_var(nfcomm%ncid, nfcomm%varid(size(nfcomm%varid)), array, start(1:4), count(1:4) )
        err = stat - NF90_NOERR
        FD4_L1_ASSERT(err/=0,'nf90_put_var: '//trim(nf90_strerror(stat)))
        deallocate(array)
      end if
      
    end if ! io ranks

    err = 0
    
  end subroutine fd4_netcdf4_write


end module fd4_netcdf4_mod
