! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010-2012 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
! Distributed & Hierarchical Space-filling curve partitioning  !
!                                                              !
!--------------------------------------------------------------!

#include "fd4.h"

!! Space-filling curve partitioning of an $fd4_domain.
!!
!! *EXPERIMENTAL DISTRIBUTED/HIERARCHICAL VERSION*
!!
!! Supports Hilbert and Morton SFC.
!! One-dimensional partitioning algorithms from $fd4_part_1d_mod are used.
!!
!! @see hilbert_mod
!! @see fd4_part_1d_mod
!! @see fd4_domain_mod
!! @see fd4_part_sfc_mod
!! @author Matthias Lieber
module fd4_part_sfcd_mod

  use fd4_globaldef_mod
  use fd4_domain_mod
  use fd4_iter_mod
  use fd4_part_1d_mod
  use fd4_part_sfc_mod
  use fd4_mpi_mod
  use hilbert_mod
  use hilbert2d_mod
  use morton_mod
  use morton2d_mod
  use mrgrnk_mod
  use stack_mod
  use stack3_mod
  use rbtree_fd4_block_mod
  use timing_mod
  
  USE_MPI
  
  implicit none
  private
  
  INCLUDE_MPIFH

  integer(i_k), parameter, public :: FD4_PART_SFCD_WGTX_MASTER = 1
  integer(i_k), parameter, public :: FD4_PART_SFCD_WGTX_COLL = 2
  
  ! public methods
#ifdef FD4_SFC_DISTR
  public :: fd4_part_sfcd_weight
#endif

  
contains  

  include 'framework/fd4_domain.incf'
  include 'util/stack.incf'
  include 'util/stack3.incf'
  include 'util/math.incf'
  include 'framework/fd4_part_sfc.incf'


#ifdef FD4_SFC_DISTR

  !! Create a new partition using block position. Considers block weights from domain%balance%blkweights.
  !!
  !! This version only works in non-adaptive block mode and when the SFC type is never changed.
  !!
  !! $opt_wgtxmeth may be $FD4_PART_SFCD_WGTX_MASTER or $FD4_PART_SFCD_WGTX_COLL.
  !! Default is $FD4_PART_SFCD_WGTX_COLL.
  !!
  !! *EXPERIMENTAL DISTRIBUTED/HIERARCHICAL VERSION*:
  !!  - ranks are splitted into equal-sized groups of size $groupsz
  !!  - a coarse-grained partitioning for the groups is calculated with a fast and parallel heuristic
  !!  - for each group a high-quality method is used to create the final partition for each rank
  !!
  subroutine fd4_part_sfcd_weight(domain, local_blocks, method, groupsz, sfctype, &
                                  balance, err, opt_binsteps, opt_accuracy, opt_blkmap_update, opt_wgtxmeth)
    type(fd4_domain), intent(inout) :: domain                         !! the domain
    type(stack3)    , intent(inout) :: local_blocks                   !! stack3 of new local partition (differs from rank to rank)
    integer(i_k)    , intent(in)    :: method                         !! 1D partitioning method to use
    integer(i_k)    , intent(in)    :: groupsz                        !! size of groups
    integer(i_k)    , intent(in)    :: sfctype                        !! space-filling curve to use
    real(r4k)       , intent(out)   :: balance                        !! estimated balance of new partitioning (not implemented)
    integer(i_k)    , intent(out)   :: err                            !! error status: 0...ok
    integer(i_k),optional,intent(in):: opt_binsteps                   !! max. number of steps for binary search methods, default: 100
    !! target ratio from best possible load balance for binary steps estimation, *must be <= 1.0*, default: 0.99
    real(r8k),   optional,intent(in):: opt_accuracy
    logical,     optional,intent(in):: opt_blkmap_update               !! update the full block map, default: true
    integer(i_k),optional,intent(in):: opt_wgtxmeth                    !! method for exchanging block weights
    target :: domain

    integer :: rank, nprocs, nblocks
    integer(i_k) :: ib, crank, pstart, pend, oldp(2)
    real(r8k) :: maxweight, maxload
    real(r8k), pointer :: blkweights_ps(:)     ! prefix sum of sfc sorted block weights
    integer(i_k), pointer :: partition(:)      ! partition vector (start indices)
    integer(i_k) :: binsteps, reqsteps
    real(r8k) :: accuracy, eps
    type(fd4_iter) :: iter
    type(fd4_mpi), pointer :: mpi
    type(part_sfc_iter) :: sfciter
    integer(i8k) :: mpi_time
    integer(i8k) :: timing_us0, timing_us0mpi, timing_us1
    logical :: blkmap_update
    
    type(fd4_domain_balance), pointer :: bal
    real(r8k) :: localweight, localweight_ps
    real(r8k) :: sumweight, bottleneck, sumload
    integer(i_k) :: cgpart0, cgpart1, cgpart, cgpart0_new, guess, ngroups
    type(stack) :: cgborders_found
    integer :: request, status(MPI_STATUS_SIZE)
    integer(i_k) :: cgborders(4), ncgblocks
    integer(i_k) :: local_copy_start, local_copy_type
    logical :: exact_part_match
    integer(i_k) :: wgtxmeth
#ifdef FD4_SFC_DISTR_DEBUG
    integer(i_k) :: emergency_abort
#endif

    err = -1
    
    FD4_L1_ASSERT(domain%varinfo%vthres_mode,'not allowed in adaptive block mode')
    if(domain%blockcount>0 .and. sfctype/=domain%balance%last_sfctype) then
      FD4_L1_ERROR('cannot change SFC type')
    end if
    if(domain%blocks%length>0 .and. domain%balance%sfc_pstart<0) then
      FD4_L1_ERROR('domain%balance%sfc_pstart not set')
    end if

    mpi    => domain%mpi
    bal    => domain%balance
    rank    = mpi%rank
    nprocs  = mpi%nprocs
    nblocks = PRODUCT(domain%blnum(1:3))
    
    blkmap_update = .true.
    if(present(opt_blkmap_update)) blkmap_update = opt_blkmap_update
    
    wgtxmeth = FD4_PART_SFCD_WGTX_COLL
    if(present(opt_wgtxmeth)) wgtxmeth = opt_wgtxmeth
    
    if(.not. associated(domain%balance%partition)) then
      allocate(domain%balance%partition(0:domain%mpi%nprocs), stat=err)
      FD4_L1_ASSERT(err/=0,'allocate domain%balance%partition')
      err = -1
    end if
    partition => domain%balance%partition

    bal%last_sfctype = sfctype

    !! create the group communicator if first call
    if(bal%sfcd_ngroups==0) then

      if(domain%stat_enabled) call gettime_ofday(timing_us0)
      
      FD4_L1_ASSERT(groupsz<=1,'groupsz must be > 1')
      FD4_L1_ASSERT(groupsz>nprocs/2,'groupsz must be <= nprocs/2')
      bal%sfcd_groupsz = groupsz
      bal%sfcd_ngroups = nprocs / groupsz
      if( bal%sfcd_groupsz * bal%sfcd_ngroups /= nprocs ) then
        bal%sfcd_groupsz = 0
        bal%sfcd_ngroups = 0
        FD4_L1_ERROR('groupsz must be divisor of nprocs')
      end if
      
      bal%sfcd_mygroup = rank / bal%sfcd_groupsz
      call MPI_Comm_split(mpi%comm, bal%sfcd_mygroup, rank, bal%sfcd_mpi%comm, err)
      err = err - MPI_SUCCESS
      FD4_L3_ASSERT(err/=0,'MPI_Comm_split failed')
      call MPI_Comm_rank(bal%sfcd_mpi%comm, bal%sfcd_mpi%rank, err)
      call MPI_Comm_size(bal%sfcd_mpi%comm, bal%sfcd_mpi%nprocs, err)
      err = -1
      FD4_L1_ASSERT(bal%sfcd_mpi%nprocs/=groupsz,'comm has wrong size')
#ifdef FD4_SFC_DISTR_VERBOSE
      write(1000+rank,*) bal%sfcd_mygroup, bal%sfcd_mpi%rank
#endif
      bal%sfcd_mpi%comm_gc  = bal%sfcd_mpi%comm
      bal%sfcd_mpi%comm_cpl = bal%sfcd_mpi%comm
      bal%sfcd_mpi%comm_mig = bal%sfcd_mpi%comm
      bal%sfcd_mpi%lognprocs = log2i(bal%sfcd_mpi%nprocs)

      if(domain%stat_enabled) then
        call gettime_ofday(timing_us1)
        domain%stat%values(domain%balance%stat_init_us) = domain%stat%values(domain%balance%stat_init_us) + (timing_us1-timing_us0)
        timing_us0 = timing_us1
      end if

    elseif(bal%sfcd_groupsz/=groupsz) then
      FD4_L1_ERROR('changing group size not implemented')
    end if

    ngroups = bal%sfcd_ngroups

    if(domain%stat_enabled) call gettime_ofday(timing_us0)

    if(domain%blockcount == nblocks) then
      !! all blocks already allocated

      !! 1. parallel prefix sum of block weights
      !!    

      ! allocate array for local block weights (twice the required size just as buffer for next time)
      if(.not. associated(bal%sfcd_blkweights_ps)) then
        allocate(bal%sfcd_blkweights_ps(0:2*domain%blocks%length), stat=err)
        FD4_L1_ASSERT(err/=0,'allocate bal%sfcd_blkweights_ps')
        err = -1
      else
        if( ubound(bal%sfcd_blkweights_ps,1) <      domain%blocks%length  .or. &
            ubound(bal%sfcd_blkweights_ps,1) > 16 * domain%blocks%length ) then
          deallocate(bal%sfcd_blkweights_ps)
          allocate(bal%sfcd_blkweights_ps(0:2*domain%blocks%length), stat=err)
          FD4_L1_ASSERT(err/=0,'allocate bal%sfcd_blkweights_ps')
          err = -1
        end if
      end if

#ifdef FD4_SFC_DISTR_VERBOSE
      write(1000+rank,*) ubound(bal%sfcd_blkweights_ps,1), domain%blocks%length
#endif    

      ! get the local block weights and write them to bal%sfcd_blkweights_ps (ordered by SFC)
      ! and get the sum of all block weights
      localweight = 0.0_r8k
      iter%cur => rbtree_fd4_block_iter_first(iter%biter, domain%blocks)
      do while(associated(iter%cur))
        call fd4_part_sfc_c2i_qsz(domain, sfctype, iter%cur%pos, ib)
        bal%sfcd_blkweights_ps(ib-bal%sfc_pstart+1) = REAL(iter%cur%weight, r8k)
        localweight = localweight + REAL(iter%cur%weight, r8k)
        iter%cur => rbtree_fd4_block_iter_next(iter%biter)
      end do

      if(domain%stat_enabled) call gettime_ofday(timing_us0mpi)

      ! parallel (exclusive) prefix sum of local block weight sums
      call MPI_Exscan(localweight, localweight_ps, 1, MPI_R8K, MPI_SUM, mpi%comm, err)
      err = err - MPI_SUCCESS
      FD4_L3_ASSERT(err/=0,'MPI_Exscan failed')
      err = -1

      ! start sending my prefix sum result to rank-1 to guarantee consistent values
      if(rank>0) then
        call MPI_Isend(localweight_ps, 1, MPI_R8K, rank-1, 5, mpi%comm, request, err)
        err = err - MPI_SUCCESS
        FD4_L3_ASSERT(err/=0,'MPI_Isend failed')
        err = -1
      end if

      if(domain%stat_enabled) then
        call gettime_ofday(timing_us1)
        domain%stat%values(bal%stat_mpi_us) = domain%stat%values(bal%stat_mpi_us) + (timing_us1-timing_us0mpi)
      end if

      ! rank 0 MPI_Exscan result is not defined
      if(rank==0) localweight_ps = 0.0_r8k
      
#ifdef FD4_SFC_DISTR_VERBOSE
      write(1000+rank,*) 'localweight: ',localweight, '  ps: ',localweight_ps
#endif
      
      ! apply (global) prefix sum to (local) weight array, i.e. start with MPI_Exscan result
      bal%sfcd_blkweights_ps(0) = localweight_ps
      do ib=1,domain%blocks%length
        bal%sfcd_blkweights_ps(ib) = bal%sfcd_blkweights_ps(ib) + bal%sfcd_blkweights_ps(ib-1)
      end do

      ! receive the prefix sum result from rank+1 to set my last value in bal%sfcd_blkweights_ps
      if(rank<nprocs-1) then
        call MPI_Recv(bal%sfcd_blkweights_ps(domain%blocks%length), 1, MPI_R8K, rank+1, 5, mpi%comm, MPI_STATUS_IGNORE, err)
        err = err - MPI_SUCCESS
        FD4_L3_ASSERT(err/=0,'MPI_Recv failed')
        err = -1
      end if
      if(rank>0) then
        call MPI_Wait(request, MPI_STATUS_IGNORE, err)
        err = err - MPI_SUCCESS
        FD4_L3_ASSERT(err/=0,'MPI_Wait failed')
        err = -1
      end if

#ifdef FD4_SFC_DISTR_VERBOSE
      do ib=0,domain%blocks%length
        write(1000+rank,*) ib, bal%sfc_pstart+ib-1, bal%sfcd_blkweights_ps(ib)
      end do
#endif

      ! get sumweight from last global rank
      sumweight = bal%sfcd_blkweights_ps(domain%blocks%length)
      if(domain%stat_enabled) call gettime_ofday(timing_us0mpi)
      call MPI_Bcast(sumweight, 1, MPI_R8K, nprocs-1, mpi%comm, err)
      err = err - MPI_SUCCESS
      FD4_L3_ASSERT(err/=0,'MPI_Bcast failed')
      err = -1
      if(domain%stat_enabled) then
        call gettime_ofday(timing_us1)
        domain%stat%values(bal%stat_mpi_us) = domain%stat%values(bal%stat_mpi_us) + (timing_us1-timing_us0mpi)
      end if
      
#ifdef FD4_SFC_DISTR_VERBOSE
      write(1000+rank,*) 'sumweight: ',sumweight
#endif

      if(domain%stat_enabled) then
        call gettime_ofday(timing_us1)
        domain%stat%values(domain%balance%stat_pfxs_us) = domain%stat%values(domain%balance%stat_pfxs_us) + (timing_us1-timing_us0)
        timing_us0 = timing_us1
      end if


      !! 2. get the coarse-grained partition borders on the SFC which
      !!    reside in local part of the block weight vector

      bottleneck = sumweight / ngroups
      exact_part_match = .false.

      ! first coarse-grained partition to scan for
      cgpart0 = CEILING(bal%sfcd_blkweights_ps(0) / bottleneck)
      cgpart0 = MAX(1,cgpart0)

      ! last coarse-grained partition to scan for
      cgpart1 = FLOOR(bal%sfcd_blkweights_ps(domain%blocks%length) / bottleneck)
      cgpart1 = cgpart1 + 1
      cgpart1 = MIN(ngroups-1,cgpart1)

#ifdef FD4_SFC_DISTR_VERBOSE
      write(1000+rank,*) 'coarse scan ', cgpart0, cgpart1
      write(1000+rank,*) 'coarse scan ', bal%sfcd_blkweights_ps(0), bal%sfcd_blkweights_ps(domain%blocks%length)
      write(1000+rank,'(A,E50.40)') 'coarse scan ', bal%sfcd_blkweights_ps(0) / bottleneck
      write(1000+rank,'(A,E50.40)') 'coarse scan ', bal%sfcd_blkweights_ps(domain%blocks%length) / bottleneck
#endif

      FD4_L1_ASSERT(cgpart1-MAX(1,cgpart0-1)<0,'cgpart bug')

      ! search with H1 for coarse-grained partitions
      cgpart0_new = HUGE(cgpart0_new)
      pstart = 0
      guess = 0
      call stack_create(cgborders_found, cgpart1-cgpart0+1)

      do cgpart = MAX(1,cgpart0-1), cgpart1
        sumload = cgpart * bottleneck
        ! pend = last block of current partition
        pend = fd4_part_1d_scan(domain%blocks%length, bal%sfcd_blkweights_ps, pstart, sumload, guess)
#ifdef FD4_SFC_DISTR_VERBOSE
        write(1000+rank,*) 'coarse scan for: ', cgpart, sumload, '->', pend+bal%sfc_pstart-1, &
                           sumload>=bal%sfcd_blkweights_ps(pend), &
                           pend==0,sumload==bal%sfcd_blkweights_ps(pend),pend<domain%blocks%length
#endif
        if(sumload>=bal%sfcd_blkweights_ps(pend)) then
          if(.not. exact_part_match .and. pend==0 .and. sumload==bal%sfcd_blkweights_ps(pend)) then
#ifdef FD4_SFC_DISTR_VERBOSE
            write(1000+rank,*) ' exact part match', cgpart, sumload, '->', pend+bal%sfc_pstart-1
#endif
            ! if the previous rank has the same match at its last block
            ! discard this match
            exact_part_match = .true.
            cgpart0_new = cgpart+1
          elseif(pend<domain%blocks%length .or. sumload==bal%sfcd_blkweights_ps(pend)) then
#ifdef FD4_SFC_DISTR_VERBOSE
            write(1000+rank,*) ' valid match     ', cgpart, sumload, '->', pend+bal%sfc_pstart-1
#endif
            call stack_push(cgborders_found, pend+bal%sfc_pstart-1)
            cgpart0_new = MIN(cgpart, cgpart0_new)
          end if
        end if
        pstart = pend
      end do

      if(cgpart0_new<HUGE(cgpart0_new)) cgpart0 = cgpart0_new
      cgpart1 = cgpart0 + cgborders_found%pos - 1

#ifdef FD4_SFC_DISTR_VERBOSE
      write(1000+rank,*) 'coarse scan update', cgpart0, cgpart1
      write(1000+rank,*) 'exact_part_match  ', exact_part_match
#endif


      if(domain%stat_enabled) then
        call gettime_ofday(timing_us1)
        domain%stat%values(bal%stat_p1dc_us) = domain%stat%values(bal%stat_p1dc_us) + (timing_us1-timing_us0)
        timing_us0 = timing_us1
        timing_us0mpi = timing_us1
      end if

#ifdef FD4_SFC_DISTR_DEBUG
      call MPI_Allreduce(cgborders_found%pos, emergency_abort, 1, MPI_I_K, MPI_SUM, mpi%comm, err)
      if(emergency_abort/=ngroups-1) then
        if(rank==0) then
          write(1000+rank,*) '!! ',emergency_abort,' borders found !!'
          write(*,*) '!! ',emergency_abort,' borders found !!'
        end if
        call mpi_finalize(err)
        stop
      end if
#endif


      !! 3. tell all groups the borders of their coarse-grained partition
      !!    (border finder sends to groups master, master bcasts to all groups members)

      ! reset the stack of requests
      call stack_create(mpi%requests, 2*(cgpart1-cgpart0+1))

      ! loop over borders and send them to resp. group master
      ib = 1
      local_copy_start = 0
      local_copy_type = 0
      cgborders(:) = -1
      do cgpart = cgpart0,cgpart1  
        ! send partition border to ranks cgpart*groupsz and (cgpart-1)*groupsz
        ! with a tag to identify lower and upper borders
        ! lower border for cgpart, tag = 10
        if(cgpart*groupsz==rank) then
#ifdef FD4_SFC_DISTR_VERBOSE
          write(1000+rank,*) 'copy lower border: group/rank ',cgpart, cgpart*groupsz
#endif
          ! do not send msg to myself, do local copy
          cgborders(1) = cgborders_found%a(ib) + 1
          cgborders(3) = rank
        else
#ifdef FD4_SFC_DISTR_VERBOSE
          write(1000+rank,*) 'send lower border: group/rank ',cgpart, cgpart*groupsz
#endif
          ! send msg
          call MPI_Isend(cgborders_found%a(ib), 1, MPI_I_K,  cgpart   *groupsz, 10, mpi%comm, request, err)
          err = err - MPI_SUCCESS
          FD4_L3_ASSERT(err/=0,'MPI_Isend failed')
          call stack_push_fast(mpi%requests, request)
        end if
        ! upper border for cgpart-1, tag = 11
        if((cgpart-1)*groupsz==rank) then
#ifdef FD4_SFC_DISTR_VERBOSE
           write(1000+rank,*) 'copy upper border: group/rank ',cgpart-1, (cgpart-1)*groupsz
#endif
          ! do not send msg to myself, do local copy
          cgborders(2) = cgborders_found%a(ib)
          cgborders(4) = rank
        else
#ifdef FD4_SFC_DISTR_VERBOSE
           write(1000+rank,*) 'send upper border: group/rank ',cgpart-1, (cgpart-1)*groupsz
#endif
          call MPI_Isend(cgborders_found%a(ib), 1, MPI_I_K, (cgpart-1)*groupsz, 11, mpi%comm, request, err)
          err = err - MPI_SUCCESS
          FD4_L3_ASSERT(err/=0,'MPI_Isend failed')
          call stack_push_fast(mpi%requests, request)
        end if
        ib = ib + 1
      end do

      ! group masters: receive partition borders
      if(bal%sfcd_mpi%rank == 0) then
      
        ! lower border
        if(cgborders(1)<0) then
          if(bal%sfcd_mygroup>0) then
            ! receive lower border, tag = 10
            call MPI_Recv(cgborders(1), 1, MPI_I_K, MPI_ANY_SOURCE, 10, mpi%comm, status, err)
            err = err - MPI_SUCCESS
            FD4_L3_ASSERT(err/=0,'MPI_Recv failed')
            ! increase by one since the upper bound of previous cgpart is sent
            cgborders(1) = cgborders(1) + 1
            cgborders(3) = status(MPI_SOURCE)
#ifdef FD4_SFC_DISTR_VERBOSE
            write(1000+rank,*) 'recv lower border: rank ',status(MPI_SOURCE)
#endif
          else
            cgborders(1) = 1
            cgborders(3) = 0
          end if
        end if
        
        ! find the rank who owns the lower border (should be cgborders(3) or a bit higher)
        do while(partition(cgborders(3)+1)<=cgborders(1))
          cgborders(3) = cgborders(3) + 1
          if(cgborders(3)>=nprocs) exit  ! empty coarse partition
        end do

        ! upper border
        if(cgborders(2)<0) then
          if(bal%sfcd_mygroup<ngroups-1) then
            ! receive upper border, tag = 11
            call MPI_Recv(cgborders(2), 1, MPI_I_K, MPI_ANY_SOURCE, 11, mpi%comm, status, err)
            err = err - MPI_SUCCESS
            FD4_L3_ASSERT(err/=0,'MPI_Recv failed')
            cgborders(4) = status(MPI_SOURCE)
#ifdef FD4_SFC_DISTR_VERBOSE
            write(1000+rank,*) 'recv upper border: rank ',status(MPI_SOURCE)
#endif
          else
            cgborders(2) = nblocks
            cgborders(4) = nprocs - 1
          end if
        end if

        ! find the rank who owns the upper border (should be cgborders(4) or a bit lower)
        do while(partition(cgborders(4))>cgborders(2))
          cgborders(4) = cgborders(4) - 1
          if(cgborders(4)<0) exit  ! empty coarse partition
        end do

      end if

      ! bcast partition borders to groups members
      call MPI_Bcast(cgborders(1), 4, MPI_I_K, 0, bal%sfcd_mpi%comm, err)
      err = err - MPI_SUCCESS
      FD4_L3_ASSERT(err/=0,'MPI_Bcast failed')

#ifdef FD4_SFC_DISTR_VERBOSE
      write(1000+rank,*) 'coarse borders: ',cgborders(1:2)
      write(1000+rank,*) 'coarse border ranks: ',cgborders(3:4)
#endif

      ! number of blocks in coarse grained partition of local group
      ncgblocks = cgborders(2) - cgborders(1) + 1
      
      if(ncgblocks==0 .and. bal%sfcd_mpi%rank==0) then
        FD4_L2_INFO('empty coarse partition')
      end if
      
      FD4_L1_ASSERT(cgborders(3)>=nprocs.and.ncgblocks>0,'lo cgrank')
      FD4_L1_ASSERT(cgborders(4)<0.and.ncgblocks>0,'up cgrank')

      if(mpi%requests%pos>0) then
        call MPI_Waitall(mpi%requests%pos, mpi%requests%a, MPI_STATUSES_IGNORE, err)
        err = err - MPI_SUCCESS
        FD4_L3_ASSERT(err/=0,'MPI_Waitall failed')
      end if    

      if(domain%stat_enabled) then
        call gettime_ofday(timing_us1)
        domain%stat%values(bal%stat_mpi_us) = domain%stat%values(bal%stat_mpi_us) + (timing_us1-timing_us0mpi)
        domain%stat%values(bal%stat_bcbr_us) = domain%stat%values(bal%stat_bcbr_us) + (timing_us1-timing_us0)
        timing_us0 = timing_us1
      end if

      err = -1


      !! 4. exchange block weights
      !!
      
      if(wgtxmeth == FD4_PART_SFCD_WGTX_MASTER) then

        call fd4_part_sfcd_exch_weights(domain, cgpart0, cgborders_found, cgborders, exact_part_match, err)
        FD4_L1_ASSERT(err/=0,'fd4_part_sfcd_exch_weights failed')

      else
      
        call fd4_part_sfcd_exch_weights_coll(domain, cgpart0, cgborders_found, cgborders, exact_part_match, err)
        FD4_L1_ASSERT(err/=0,'fd4_part_sfcd_exch_weights_coll failed')

      end if

      blkweights_ps => domain%balance%blkweights_ps

      call stack_delete(cgborders_found)

      if(domain%stat_enabled) then
        call gettime_ofday(timing_us1)
        domain%stat%values(domain%balance%stat_exwgt_us) = domain%stat%values(domain%balance%stat_exwgt_us) &
                                                         + (timing_us1-timing_us0)
        timing_us0 = timing_us1
      end if

#ifdef FD4_SFC_DISTR_VERBOSE
      if(ncgblocks>0) then
        do ib=0,ncgblocks
          write(1000+rank,*) ib, ib+cgborders(1)-1, blkweights_ps(ib)
        end do
      end if
#endif


      !! 5. Create high-quality partitioning within group
      !!

      ! save old local partition
      oldp(1) = bal%sfc_pstart
      oldp(2) = oldp(1) + domain%blocks%length - 1

      ! save old partition vector
      if(.not. blkmap_update) then
        if(.not. associated(bal%partition_old)) then
          allocate(bal%partition_old(0:domain%mpi%nprocs), stat=err)
          FD4_L1_ASSERT(err/=0,'allocate bal%partition_old')
          err = -1
        end if
        bal%partition => bal%partition_old
        bal%partition_old => partition
        partition => bal%partition
      end if

      ! prefix sum of group's blocks:  blkweights_ps(0) ... blkweights_ps(ncgblocks)

      if(ncgblocks>0) then

        ! get max block weight in group
        maxweight=blkweights_ps(1)-blkweights_ps(0)
        do ib=2,ncgblocks
          maxweight = MAX(blkweights_ps(ib)-blkweights_ps(ib-1), maxweight)
        end do

        if(method==FD4_PART_SFC_BINSRCH .or. method==FD4_PART_SFC_PARALLEL) then
          accuracy = 0.99_r8k
          binsteps = 100
          if(present(opt_accuracy)) accuracy=MAX(MIN(opt_accuracy,1.0_r8k),0.1_r8k)
          if(present(opt_binsteps)) binsteps=MAX(MIN(opt_binsteps,10000),1)
          ! compute the bottleneck error value which is necessary to get
          !  (accuracy * optimal_balance)
          ! as the resulting loadbalance of the 1d partitioning
          !  (eps is 0 if accuracy is 1.0: in this case the methods is exact if binsteps is large enough)
          eps = (1.0_r8k-accuracy) * blkweights_ps(ncgblocks) / (accuracy*groupsz)
          ! this would be the necessary number of binary search steps for the given accuracy:
          !steps = log2i( CEILING( maxweight / eps, i_k ) )
        end if

        !! call partitioning routine

        if(method==FD4_PART_SFC_BINSRCH) then
          ! good serial method
          call fd4_part_1d_binsrch(ncgblocks, groupsz, blkweights_ps(0:ncgblocks), maxweight, eps, binsteps, &
                                           partition, maxload, reqsteps)
          if(domain%stat_enabled) then
            domain%stat%values(bal%stat_binsteps) = MAX(domain%stat%values(bal%stat_binsteps), INT(reqsteps,i8k))
          end if
          !
        elseif(method==FD4_PART_SFC_PARALLEL) then    
          ! parallel method
          call fd4_part_1d_parallel(ncgblocks, groupsz, blkweights_ps(0:ncgblocks), maxweight, eps, binsteps, &
                                            bal%sfcd_mpi, partition, maxload, reqsteps, mpi_time, err)
          FD4_L1_ASSERT(err/=0,'fd4_part_1d_parallel failed')
          err = -1
          if(domain%stat_enabled) then
            domain%stat%values(bal%stat_binsteps) = MAX(domain%stat%values(bal%stat_binsteps), INT(reqsteps,i8k))
            domain%stat%values(bal%stat_mpi_us)   = domain%stat%values(bal%stat_mpi_us) + mpi_time
          end if
          !
        elseif(method==FD4_PART_SFC_SIMPLE) then
          ! simple method 1
          call fd4_part_1d_simple_h1(ncgblocks, groupsz, blkweights_ps(0:ncgblocks), partition, maxload)
          !
        else
          ! simple method 2
          call fd4_part_1d_simple_h2(ncgblocks, groupsz, blkweights_ps(0:ncgblocks), partition, maxload)
          !
        end if

      else ! ncgblocks <= 0

        ! no blocks in coarse grained partition (awkward, but valid)
        maxweight = 0.0_r8k
        partition(0:groupsz) = 1

      end if

      if(domain%stat_enabled) then
        call gettime_ofday(timing_us1)
        domain%stat%values(bal%stat_p1d_us) = domain%stat%values(bal%stat_p1d_us) + (timing_us1-timing_us0)
        timing_us0 = timing_us1
        timing_us0mpi = timing_us1
      end if

#ifdef FD4_SFC_DISTR_VERBOSE
      do ib=0,groupsz
        write(1000+rank,*) 'partition: ',ib+bal%sfcd_mygroup*groupsz, partition(ib)+cgborders(1)-1
      end do
#endif    

      !! 6: distribute partition vector over all ranks
      !!

      pstart = partition(bal%sfcd_mpi%rank)+cgborders(1)-1
      call MPI_Allgather(pstart, 1, MPI_I_K, partition(0), 1, MPI_I_K, mpi%comm, err)
      err = err - MPI_SUCCESS
      FD4_L3_ASSERT(err/=0,'MPI_Allgather failed')
      err = -1
      if(domain%stat_enabled) then
        call gettime_ofday(timing_us1)
        domain%stat%values(bal%stat_mpi_us) = domain%stat%values(bal%stat_mpi_us) + (timing_us1-timing_us0mpi)
        domain%stat%values(bal%stat_dspv_us) = domain%stat%values(bal%stat_dspv_us) + (timing_us1-timing_us0)
        timing_us0 = timing_us1
      end if
      partition(nprocs) = nblocks + 1

      pend   = partition(rank+1)-1
      bal%sfc_pstart = pstart
      if(pstart<=pend) then
        bal%localweight = REAL(blkweights_ps(pend-cgborders(1)+1) - blkweights_ps(pstart-cgborders(1)),r4k)
      else
        bal%localweight = 0.0
      end if

#ifdef FD4_SFC_DISTR_VERBOSE
      if(rank==0) print *,partition
#endif

    elseif(domain%blockcount == 0) then
      !! no blocks allocated yet
      
      !! create an initial partitioning
      !! TODO: consider weights (i.e. block sizes) here

      do crank=0,nprocs
        partition(crank) = INT( (INT(nblocks,i8k) * INT(crank,i8k)) / INT(nprocs,i8k) + 1_i8k, i_k )
      end do
      
      pstart = partition(rank)
      pend   = partition(rank+1)-1
      bal%sfc_pstart = pstart
      bal%localweight = 0.0
      oldp(1:2) = -1

      if(domain%stat_enabled) then
        call gettime_ofday(timing_us1)
        domain%stat%values(bal%stat_p1d_us) = domain%stat%values(bal%stat_p1d_us) + (timing_us1-timing_us0)
        timing_us0 = timing_us1
      end if

#ifdef FD4_SFC_DISTR_VERBOSE
      if(rank==0) print *,'initial partition:'
      if(rank==0) print *,partition
#endif

    else
      !! adaptive block mode
      FD4_L1_ERROR('not allowed in adaptive block mode')
    end if


    !! finally apply partition to FD4
    !!

    err = -1
    balance = -1

    domain%blkmap_last = domain%blkmap_now
    domain%blkmap_now  = 3 - domain%blkmap_now

    ! allocate local list of blocks
    if(pstart<=pend) then
      call stack3_create(local_blocks, pend - pstart + 1)
      local_blocks%pos = pend - pstart + 1
    else
      local_blocks%pos = 0
    end if

#ifdef FD4_SFC_DISTR_VERBOSE
      write(1000+rank,*) 'pstart ',pstart
      write(1000+rank,*) 'pend   ',pend
      write(1000+rank,*) 'local_blocks%pos   ',local_blocks%pos
#endif    

    call fd4_part_sfc_iter_init(sfciter, domain, sfctype)

    ! set local list of blocks
    if(pstart<=pend) then
      call fd4_part_sfc_iter_set(sfciter, pstart)
      do while(sfciter%idx>0 .and. sfciter%idx<=pend)
        local_blocks%a(1:3,sfciter%idx-pstart+1) = sfciter%p(1:3)
        call fd4_part_sfc_iter_next(sfciter)
      end do
    end if

    bal%blkmap_needs_update = .true.
    
    ! here is the actual blkmap update
    if(blkmap_update) then
      ! full update
      call fd4_part_sfc_update_blkmap(domain)
    else
      ! update only the necessary blocks
      call fd4_part_sfc_update_blkmap_loc(domain, sfciter, oldp, pstart, pend, local_blocks, err)
      FD4_L1_ASSERT(err/=0,'fd4_part_sfc_update_blkmap_loc failed')
      err = -1
    end if

    if(domain%stat_enabled) then
      call gettime_ofday(timing_us1)
      domain%stat%values(domain%balance%stat_blkm_us) = domain%stat%values(domain%balance%stat_blkm_us) + (timing_us1-timing_us0)
      ! fill statistics
      domain%stat%sfcbnd(1) = MIN(domain%stat%sfcbnd(1), pstart)
      domain%stat%sfcbnd(2) = MAX(domain%stat%sfcbnd(2), pend)
    end if

#if ( FD4_VERBOSE_LEVEL >= 4 )
    write(fd4_out,'(A,A,I6,E13.4)') FD4_LINEHD,'new partition: ',local_blocks%pos, domain%balance%localweight
#endif
  
    err = 0

  end subroutine fd4_part_sfcd_weight


  !! Exchange block weights.
  !! This is $FD4_PART_SFCD_WGTX_MASTER.
  !!
  !! Method:
  !! - weights are sent to group master with p2p messages
  !! - master bcasts to group members
  !!
  subroutine fd4_part_sfcd_exch_weights(domain, cgpart0, cgborders_found, cgborders, exact_part_match, err)
    type(fd4_domain), intent(inout) :: domain              !! the domain
    integer(i_k)    , intent(in)    :: cgpart0             !! lower bound for search of local coarse border matches
    type(stack)     , intent(in)    :: cgborders_found     !! stack containing local matches of coarse partition borders
    integer(i_k)    , intent(in)    :: cgborders(4)        !! partitions borders/ranks of own coarse partition
    logical         , intent(in)    :: exact_part_match    !! first local block exactly starts new coarse partition?
    integer(i_k)    , intent(out)   :: err                 !! error status: 0...ok
    target :: domain
    
    integer :: rank, nprocs
    integer(i_k) :: ib, crank
    real(r8k), pointer :: blkweights_ps(:)     ! prefix sum of sfc sorted block weights
    integer(i_k), pointer :: partition(:)      ! partition vector (start indices)
    type(fd4_mpi), pointer :: mpi
    integer(i8k) :: timing_us0mpi, timing_us1
    type(fd4_domain_balance), pointer :: bal
    integer(i_k) :: cgpart1, cgpart
    integer :: request, tag
    integer(i_k) :: ncgblocks, start_pos, num_elem, groupsz
    integer(i_k) :: local_copy_start, local_copy_num
    logical :: first_rank_in_cgpart
    ! add this to message tag so that overlap with lower tags not possible
    integer, parameter :: mintag = 20
#ifdef FD4_SFC_DISTR_DEBUG
    integer(i_k) :: emergency_abort
    emergency_abort=0
#endif

    err = -1

    mpi    => domain%mpi
    bal    => domain%balance
    rank    = mpi%rank
    nprocs  = mpi%nprocs
    partition => domain%balance%partition
    groupsz = bal%sfcd_groupsz
    cgpart1 = cgpart0 + cgborders_found%pos - 1
    ncgblocks = cgborders(2) - cgborders(1) + 1
    mpi%requests%pos = 0

    ! send block weights to group masters
    ib = 1
    start_pos = 1
    local_copy_start = 0
    local_copy_num = -1
    ! owner of first block is always first rank in coarse partition
    first_rank_in_cgpart = (partition(rank)==1 .or. exact_part_match)
#ifdef FD4_SFC_DISTR_VERBOSE
    write(1000+rank,*) 'send weights: first_rank_in_cgpart ',first_rank_in_cgpart
    write(1000+rank,*) 'send weights: partition(rank)==1   ',partition(rank)==1
    write(1000+rank,*) 'send weights: exact_part_match     ',exact_part_match
#endif
    do cgpart = cgpart0-1, cgpart1
      if(cgpart<cgpart1) then
        num_elem = cgborders_found%a(ib) - bal%sfc_pstart + 1 - start_pos + 1
      else
        num_elem = domain%blocks%length - start_pos + 1
      end if
      if(num_elem>0) then
        if(first_rank_in_cgpart) then
          ! first rank in a coarse partition additionally sends previous block
          start_pos = start_pos - 1
          num_elem = num_elem + 1
          first_rank_in_cgpart = .false.
        end if
#ifdef FD4_SFC_DISTR_VERBOSE
        write(1000+rank,*) 'send weights: group ',cgpart
        write(1000+rank,*) 'send weights: l pos ',start_pos, start_pos+num_elem-1
        write(1000+rank,*) 'send weights: g pos ',start_pos+bal%sfc_pstart-1, start_pos+bal%sfc_pstart-1+num_elem-1
#endif
        crank = cgpart*groupsz
        if(crank == rank) then
          ! do not send msg to myself, prepare local copy
          local_copy_start = start_pos
          local_copy_num = num_elem
        else
          ! go! use msg size as tag to force deadlocks if msg size differs at sender and receiver
          if(domain%stat_enabled) call gettime_ofday(timing_us0mpi)
          tag = mintag + INT(num_elem)
          call MPI_Isend(bal%sfcd_blkweights_ps(start_pos), num_elem, MPI_R8K, crank, &
                          tag, mpi%comm, request, err)
          err = err - MPI_SUCCESS
          FD4_L3_ASSERT(err/=0,'MPI_Isend failed')
          call MPI_Request_free(request, err)
          err = -1
          if(domain%stat_enabled) then
            call gettime_ofday(timing_us1)
            domain%stat%values(bal%stat_mpi_us) = domain%stat%values(bal%stat_mpi_us) + (timing_us1-timing_us0mpi)
          end if
        end if
      end if
      if(cgpart<cgpart1) start_pos = start_pos + num_elem
      if(start_pos>domain%blocks%length) exit
      first_rank_in_cgpart = .true.
      ib = ib + 1
    end do

#ifdef FD4_SFC_DISTR_VERBOSE
    if(rank==0) print *,partition
#endif

    ! allocate vector for all block weights of the local group
    if(.not. associated(bal%blkweights_ps)) then
      allocate(bal%blkweights_ps(0:MAX(2*ncgblocks,16)), stat=err)
      FD4_L1_ASSERT(err/=0,'allocate bal%blkweights_ps')
      err = -1
    else
      if( ubound(bal%blkweights_ps,1) <      ncgblocks  .or. &
          ubound(bal%blkweights_ps,1) > 16 * ncgblocks ) then
        deallocate(bal%blkweights_ps)
        allocate(bal%blkweights_ps(0:MAX(2*ncgblocks,16)), stat=err)
        FD4_L1_ASSERT(err/=0,'allocate bal%blkweights_ps')
        err = -1
      end if
    end if
    blkweights_ps => bal%blkweights_ps


    ! group masters receive block weights
    if( bal%sfcd_mpi%rank == 0 .and. ncgblocks>0 ) then

      ! reset the stack of requests
      call stack_create(mpi%requests, cgborders(4)-cgborders(3)+1)
      first_rank_in_cgpart = .true.
      start_pos = cgborders(1)
      crank = cgborders(3)
#ifdef FD4_SFC_DISTR_DEBUG
      ! find rank which owns block at start_pos
      call fd4_part_sfc_find_partition(partition, start_pos, crank)
      if(crank/=cgborders(3)) then
        write(1000+rank,*) 'EMERGENCY ABORT 2'
        write(1000+rank,*) ' crank: ',crank
        write(1000+rank,*) ' cgborders(3): ',cgborders(3)
        emergency_abort=1
      end if
#endif
      if(domain%stat_enabled) call gettime_ofday(timing_us0mpi)
      ! loop over ranks to recv from
      do
        num_elem = MIN(partition(crank+1) - start_pos, cgborders(2) - start_pos + 1)
        if(num_elem>0) then
          if(first_rank_in_cgpart) then
            ! first rank in a coarse partition additionally sends previous block
            start_pos = start_pos - 1
            num_elem = num_elem + 1
#ifndef FD4_SFC_DISTR_DEBUG
            first_rank_in_cgpart = .false.
#endif
          end if
#ifdef FD4_SFC_DISTR_VERBOSE
          write(1000+rank,*) 'recv weights: rank ',crank
          write(1000+rank,*) 'recv weights: l pos ',start_pos-cgborders(1)+1, start_pos-cgborders(1)+1+num_elem-1
          write(1000+rank,*) 'recv weights: g pos ',start_pos, start_pos+num_elem-1
#endif
          if(crank == rank) then
            ! copy from own local weight vector
#ifdef FD4_SFC_DISTR_DEBUG
  if(num_elem/=local_copy_num) then
  !if(.true.) then
    write(1000+rank,*) 'EMERGENCY ABORT'
    write(1000+rank,*) 'sfcd group/rank: ',bal%sfcd_mygroup, bal%sfcd_mpi%rank
    write(1000+rank,*) 'partition:'
    do ib=0,nprocs
      write(1000+rank,*) ib,partition(ib)
    end do
    write(1000+rank,*) 'local block weights:'
    do ib=0,domain%blocks%length
      write(1000+rank,*) ib, bal%sfc_pstart+ib-1, bal%sfcd_blkweights_ps(ib)
    end do
    write(1000+rank,*) 'coarse scan ', cgpart0, cgpart1
    write(1000+rank,*) 'cgborders:',cgborders(1:2)
    do ib=1,cgborders_found%pos
      write(1000+rank,*) 'cgborders_found: ',ib,cgborders_found%a(ib)
    end do
    write(1000+rank,*) 'num_elem:',num_elem
    write(1000+rank,*) 'local_copy_num:',local_copy_num
    write(1000+rank,*) 'local_copy_start:',local_copy_start
    write(1000+rank,*) 'first_rank_in_cgpart:',first_rank_in_cgpart
    write(1000+rank,*) 'recv weights: rank ',crank
    write(1000+rank,*) 'recv weights: l pos ',start_pos-cgborders(1)+1, start_pos-cgborders(1)+1+num_elem-1
    write(1000+rank,*) 'recv weights: g pos ',start_pos, start_pos+num_elem-1
    !
    ib = 1
    start_pos = 1
    first_rank_in_cgpart = (partition(rank)==1 .or. exact_part_match)
    do cgpart = cgpart0-1, cgpart1
      if(cgpart<cgpart1) then
        num_elem = cgborders_found%a(ib) - bal%sfc_pstart + 1 - start_pos + 1
      else
        num_elem = domain%blocks%length - start_pos + 1
      end if
      if(num_elem>0) then
        if(first_rank_in_cgpart) then
          ! first rank in a coarse partition additionally sends previous block
          start_pos = start_pos - 1
          num_elem = num_elem + 1
        end if
        write(1000+rank,*) 'send weights: group ',cgpart
        write(1000+rank,*) 'send weights: rank  ',cgpart*groupsz
        write(1000+rank,*) 'send weights: first ',first_rank_in_cgpart, exact_part_match
        write(1000+rank,*) 'send weights: l pos ',start_pos, start_pos+num_elem-1
        write(1000+rank,*) 'send weights: g pos ',start_pos+bal%sfc_pstart-1, start_pos+bal%sfc_pstart-1+num_elem-1
        first_rank_in_cgpart = .false.
      end if
      if(cgpart<cgpart1) start_pos = start_pos + num_elem
      if(start_pos>domain%blocks%length) exit
      first_rank_in_cgpart = .true.
      ib = ib + 1
    end do
    FD4_L2_INFO('num_elem/=local_copy_num')
    emergency_abort=1
    exit
  end if
#else
            FD4_L1_ASSERT(num_elem/=local_copy_num,'num_elem/=local_copy_num')
#endif
            blkweights_ps(start_pos-cgborders(1)+1:start_pos-cgborders(1)+num_elem) = &
              bal%sfcd_blkweights_ps(local_copy_start:local_copy_start+num_elem-1)
          else
            ! go! use msg size as tag to force deadlocks if msg size differs at sender and receiver
            tag = mintag + INT(num_elem)
            call MPI_Irecv(blkweights_ps(start_pos-cgborders(1)+1), num_elem, MPI_R8K, crank, &
                                         tag, mpi%comm, request, err)
            err = err - MPI_SUCCESS
            FD4_L3_ASSERT(err/=0,'MPI_Irecv failed')
            err = -1
            call stack_push_fast(mpi%requests, request)
          end if
        end if
#ifdef FD4_SFC_DISTR_DEBUG
        first_rank_in_cgpart = .false.
#endif
        if(num_elem == cgborders(2) - start_pos + 1) exit
        start_pos = start_pos + num_elem
        crank = crank + 1
        FD4_L1_ASSERT(crank>=nprocs,'bug in recv loop')
      end do
      if(domain%stat_enabled) then
        call gettime_ofday(timing_us1)
        domain%stat%values(bal%stat_mpi_us) = domain%stat%values(bal%stat_mpi_us) + (timing_us1-timing_us0mpi)
      end if

      if(crank/=cgborders(4)) then
#ifdef FD4_SFC_DISTR_DEBUG
        write(1000+rank,*) 'EMERGENCY ABORT 2'
        write(1000+rank,*) ' crank: ',crank
        write(1000+rank,*) ' cgborders(4): ',cgborders(4)
        emergency_abort=1
#else
        FD4_L1_ERROR('!rank2: '//i2s(crank-1)//i2s(cgborders(4)))
#endif
      end if

    end if ! group masters receive block weights

#ifdef FD4_SFC_DISTR_DEBUG
    call MPI_Allreduce(MPI_IN_PLACE, emergency_abort, 1, MPI_I_K, MPI_MAX, mpi%comm, err)
    if(emergency_abort>0) then
      err = -42
      return
    end if
#endif

    ! wait for recv to complete
    if(mpi%requests%pos>0) then
      if(domain%stat_enabled) call gettime_ofday(timing_us0mpi)
      call MPI_Waitall(mpi%requests%pos, mpi%requests%a, MPI_STATUSES_IGNORE, err)
      err = err - MPI_SUCCESS
      FD4_L3_ASSERT(err/=0,'MPI_Waitall failed')
      err = -1
      if(domain%stat_enabled) then
        call gettime_ofday(timing_us1)
        domain%stat%values(bal%stat_mpi_us) = domain%stat%values(bal%stat_mpi_us) + (timing_us1-timing_us0mpi)
      end if
    end if

    ! group master creates local prefix sum (i.e. substract first value from all)
    if( bal%sfcd_mpi%rank == 0 .and. ncgblocks>0 ) then
      blkweights_ps(1:ncgblocks) = blkweights_ps(1:ncgblocks) - blkweights_ps(0)
    end if
    blkweights_ps(0) = 0.0_r8k

    ! group master bcasts weights to all group members
    if( ncgblocks>0 ) then
      if(domain%stat_enabled) call gettime_ofday(timing_us0mpi)
      call MPI_Bcast(blkweights_ps(1), ncgblocks, MPI_R8K, 0, bal%sfcd_mpi%comm, err)
      err = err - MPI_SUCCESS
      FD4_L3_ASSERT(err/=0,'MPI_Bcast failed')
      err = -1
      if(domain%stat_enabled) then
        call gettime_ofday(timing_us1)
        domain%stat%values(bal%stat_mpi_us) = domain%stat%values(bal%stat_mpi_us) + (timing_us1-timing_us0mpi)
      end if
    end if

    err = 0

  end subroutine fd4_part_sfcd_exch_weights


  !! Exchange block weights. Less p2p traffic than $fd4_part_sfcd_exch_weights.
  !! This is $FD4_PART_SFCD_WGTX_COLL.
  !!
  !! Method:
  !! - weights from non-group members are sent to first/last group member with p2p messages
  !! - weights are exchanged within group with MPI_Allgatherv (in-place)
  !!
  subroutine fd4_part_sfcd_exch_weights_coll(domain, cgpart0, cgborders_found, cgborders, exact_part_match, err)
    type(fd4_domain), intent(inout) :: domain              !! the domain
    integer(i_k)    , intent(in)    :: cgpart0             !! lower bound for search of local coarse border matches
    type(stack)     , intent(in)    :: cgborders_found     !! stack containing local matches of coarse partition borders
    integer(i_k)    , intent(in)    :: cgborders(4)        !! partitions borders/ranks of own coarse partition
    logical         , intent(in)    :: exact_part_match    !! first local block exactly starts new coarse partition?
    integer(i_k)    , intent(out)   :: err                 !! error status: 0...ok
    target :: domain
    
    integer :: rank, nprocs
    integer(i_k) :: ib, crank, grank
    real(r8k), pointer :: blkweights_ps(:)     ! prefix sum of sfc sorted block weights
    integer(i_k), pointer :: partition(:)      ! partition vector (start indices)
    type(fd4_mpi), pointer :: mpi
    integer(i8k) :: timing_us0mpi, timing_us1
    type(fd4_domain_balance), pointer :: bal
    integer(i_k) :: cgpart1, cgpart
    integer :: request, tag
    integer(i_k) :: ncgblocks, start_pos, num_elem, groupsz, rankloop(2)
    integer(i_k) :: local_copy_start, local_copy_num
    logical :: first_rank_in_cgpart
    ! add this to message tag so that overlap with lower tags not possible
    integer, parameter :: mintag = 20
    logical :: cgborders_outside_group
#ifdef FD4_SFC_DISTR_COMP_WGT
    real(r8k), target :: blkweights_ps_debug(0:8192*2)
#endif

!#define FD4_SFC_DISTR_VERBOSE_SYNC
#ifdef FD4_SFC_DISTR_VERBOSE_SYNC
    integer :: ferr
    ! Declare the interface for POSIX fsync function
!    interface
!      function fsync (fd) bind(c,name="fsync")
!      use iso_c_binding, only: c_int
!        integer(c_int), value :: fd
!        integer(c_int) :: fsync
!      end function fsync
!    end interface
#endif

    err = -1

    mpi    => domain%mpi
    bal    => domain%balance
    rank    = mpi%rank
    nprocs  = mpi%nprocs
    partition => domain%balance%partition
    groupsz = bal%sfcd_groupsz
    cgpart1 = cgpart0 + cgborders_found%pos - 1
    ncgblocks = cgborders(2) - cgborders(1) + 1
    mpi%requests%pos = 0

    call fd4_mpi_realloc_array(mpi%recvcnt, groupsz, 0, err)
    FD4_L1_ASSERT(err/=0,'fd4_mpi_realloc_array failed')
    call fd4_mpi_realloc_array(mpi%recvdis, groupsz, 0, err)
    FD4_L1_ASSERT(err/=0,'fd4_mpi_realloc_array failed')

    local_copy_start = 0
    local_copy_num = 0

    ! send block weights to first/last group members of other groups
    if(domain%blocks%length>0) then
    
      ib = 1
      start_pos = 1
      ! owner of first block is always first rank in coarse partition
      first_rank_in_cgpart = (partition(rank)==1 .or. exact_part_match)
#ifdef FD4_SFC_DISTR_VERBOSE
      write(1000+rank,*) 'C send weights: first_rank_in_cgpart ',first_rank_in_cgpart
      write(1000+rank,*) 'C send weights: partition(rank)==1   ',partition(rank)==1
      write(1000+rank,*) 'C send weights: exact_part_match     ',exact_part_match
#endif

      if(domain%stat_enabled) call gettime_ofday(timing_us0mpi)

      do cgpart = cgpart0-1, cgpart1
        if(cgpart<cgpart1) then
          num_elem = cgborders_found%a(ib) - bal%sfc_pstart + 1 - start_pos + 1
        else
          num_elem = domain%blocks%length - start_pos + 1
        end if
        if(num_elem>0) then
          if(first_rank_in_cgpart) then
            ! first rank in a coarse partition additionally sends previous block
            start_pos = start_pos - 1
            num_elem = num_elem + 1
            first_rank_in_cgpart = .false.
          end if
#ifdef FD4_SFC_DISTR_VERBOSE
          write(1000+rank,*) 'C send weights: group ',cgpart
          write(1000+rank,*) 'C send weights: l pos ',start_pos, start_pos+num_elem-1
          write(1000+rank,*) 'C send weights: g pos ',start_pos+bal%sfc_pstart-1, start_pos+bal%sfc_pstart-1+num_elem-1
#endif
          if(cgpart==bal%sfcd_mygroup) then
            ! this part of my blocks stays within my group: prepare local copy for allgatherv
#ifdef FD4_SFC_DISTR_VERBOSE
            write(1000+rank,*) 'C send weights: prepare local copy'
#endif
            ! prepare local copy
            local_copy_start = start_pos
            local_copy_num = num_elem
          else
            crank = cgpart*groupsz
            ! send to last group member my group comes after group to send to
            if(cgpart<bal%sfcd_mygroup) crank = crank + groupsz-1
#ifdef FD4_SFC_DISTR_VERBOSE
            write(1000+rank,*) 'C send weights:  p2p to rank ', crank, '   size ',num_elem
#endif
            ! go! use msg size as tag to force deadlocks if msg size differs at sender and receiver
            tag = mintag + INT(num_elem)
            call MPI_Isend(bal%sfcd_blkweights_ps(start_pos), num_elem, MPI_R8K, crank, &
                            tag, mpi%comm, request, err)
            err = err - MPI_SUCCESS
            FD4_L3_ASSERT(err/=0,'MPI_Isend failed')
            call MPI_Request_free(request, err)
            err = -1
          end if
        end if ! num_elem>0
        if(cgpart<cgpart1) start_pos = start_pos + num_elem
        if(start_pos>domain%blocks%length) exit
        first_rank_in_cgpart = .true.
        ib = ib + 1
      end do

      if(domain%stat_enabled) then
        call gettime_ofday(timing_us1)
        domain%stat%values(bal%stat_mpi_us) = domain%stat%values(bal%stat_mpi_us) + (timing_us1-timing_us0mpi)
      end if

    end if ! domain%blocks%length>0

#ifdef FD4_SFC_DISTR_VERBOSE
    if(rank==0) print *,partition
#endif

    ! allocate vector for all block weights of the local group
    if(.not. associated(bal%blkweights_ps)) then
      allocate(bal%blkweights_ps(0:MAX(2*ncgblocks,16)), stat=err)
      FD4_L1_ASSERT(err/=0,'allocate bal%blkweights_ps')
      err = -1
    else
      if( ubound(bal%blkweights_ps,1) <      ncgblocks  .or. &
          ubound(bal%blkweights_ps,1) > 16 * ncgblocks ) then
        deallocate(bal%blkweights_ps)
        allocate(bal%blkweights_ps(0:MAX(2*ncgblocks,16)), stat=err)
        FD4_L1_ASSERT(err/=0,'allocate bal%blkweights_ps')
        err = -1
      end if
    end if
    blkweights_ps => bal%blkweights_ps

#ifdef FD4_SFC_DISTR_COMP_WGT
    blkweights_ps => blkweights_ps_debug
    blkweights_ps_debug = -1.0_r8k
#endif    

    ! first and last group members receive block weights
    if( (bal%sfcd_mpi%rank == 0 .or. bal%sfcd_mpi%rank == groupsz-1) .and. ncgblocks>0 ) then

      if( bal%sfcd_mpi%rank == 0 ) then
        ! first group member receives from ranks before group
        rankloop(1) = cgborders(3)                         ! first rank to recv from
        rankloop(2) = MIN(cgborders(4), rank-1)            ! last  rank to recv from
        start_pos = cgborders(1)                           ! first block to recv
      else
        ! last group member receives from ranks after group
        rankloop(1) = MAX(cgborders(3), rank+1)            ! first rank to recv from
        rankloop(2) = cgborders(4)                         ! last  rank to recv from
        start_pos = MAX(cgborders(1), partition(rank+1))   ! first block to recv
      end if

#ifdef FD4_SFC_DISTR_VERBOSE
      write(1000+rank,*) 'C recv weights: rankloop ',rankloop(1:2)
      write(1000+rank,*) 'C recv weights: start_pos ',start_pos
#endif

      if(rankloop(2) >= rankloop(1)) then

        ! reset the stack of requests
        call stack_create(mpi%requests, rankloop(2)-rankloop(1)+1)
        first_rank_in_cgpart = ( start_pos == cgborders(1) )

        if(domain%stat_enabled) call gettime_ofday(timing_us0mpi)

        ! loop over ranks to recv from
        do crank = rankloop(1), rankloop(2)
          num_elem = MIN(partition(crank+1) - start_pos, cgborders(2) - start_pos + 1)
          if(num_elem>0) then
            if(first_rank_in_cgpart) then
              ! first rank in a coarse partition additionally sends previous block
              start_pos = start_pos - 1
              num_elem = num_elem + 1
              first_rank_in_cgpart = .false.
            end if
#ifdef FD4_SFC_DISTR_VERBOSE
            write(1000+rank,*) 'C recv weights: rank ',crank, '   size ',num_elem
            write(1000+rank,*) 'C recv weights: l pos ',start_pos-cgborders(1)+1, start_pos-cgborders(1)+1+num_elem-1
            write(1000+rank,*) 'C recv weights: g pos ',start_pos, start_pos+num_elem-1
#endif
            ! go! use msg size as tag to force deadlocks if msg size differs at sender and receiver
            tag = mintag + INT(num_elem)
            call MPI_Irecv(blkweights_ps(start_pos-cgborders(1)+1), num_elem, MPI_R8K, crank, &
                                         tag, mpi%comm, request, err)
            err = err - MPI_SUCCESS
            FD4_L3_ASSERT(err/=0,'MPI_Irecv failed')
            err = -1
            call stack_push_fast(mpi%requests, request)
#ifdef FD4_SFC_DISTR_COMP_WGT
            !!DEBUG: compare blocks weights with those sent by the other method
            !!DEBUG: (only if blkweights_ps is different array!)
            call MPI_Wait(request, MPI_STATUS_IGNORE, err)
            err = -1
            do ib=start_pos-cgborders(1)+2,start_pos-cgborders(1)+num_elem
              if(blkweights_ps(ib)-blkweights_ps(ib-1) /= &
                 bal%blkweights_ps(ib)-bal%blkweights_ps(ib-1)) then
                print *,rank,' recv err at',ib-1,ib+cgborders(1)-2,blkweights_ps(ib-1),bal%blkweights_ps(ib-1)
                print *,rank,' recv err at',ib,ib+cgborders(1)-1,blkweights_ps(ib),bal%blkweights_ps(ib)
                !err = -42
                !return
              end if
            end do
#endif
          end if
          start_pos = start_pos + num_elem
        end do ! rankloop

        if(domain%stat_enabled) then
          call gettime_ofday(timing_us1)
          domain%stat%values(bal%stat_mpi_us) = domain%stat%values(bal%stat_mpi_us) + (timing_us1-timing_us0mpi)
        end if

      end if ! rankloop(2) >= rankloop(1)

    end if ! first and last group members receive block weights


#ifdef FD4_SFC_DISTR_VERBOSE_SYNC
    flush(1000+rank)
    ferr=fsync(fnum(1000+rank))
#endif

    ! wait for recv to complete
    if(mpi%requests%pos>0) then
      if(domain%stat_enabled) call gettime_ofday(timing_us0mpi)
      call MPI_Waitall(mpi%requests%pos, mpi%requests%a, MPI_STATUSES_IGNORE, err)
      err = err - MPI_SUCCESS
      FD4_L3_ASSERT(err/=0,'MPI_Waitall failed')
      err = -1
      if(domain%stat_enabled) then
        call gettime_ofday(timing_us1)
        domain%stat%values(bal%stat_mpi_us) = domain%stat%values(bal%stat_mpi_us) + (timing_us1-timing_us0mpi)
      end if
    end if

    ! distribute block weights within group to all ranks with MPI_Allgatherv
    if( ncgblocks>0 ) then
    
      ! loop over ranks of group and prepare MPI_Allgatherv
      mpi%recvdis(0) = 0
      crank = bal%sfcd_mygroup*groupsz ! global rank
      do grank = 0, groupsz-1
        cgborders_outside_group = &
              (grank==0         .and. crank>cgborders(4)) .or. &  ! upper border before 1st group rank
              (grank==groupsz-1 .and. crank<cgborders(3))         ! lower border after last group rank
#ifdef FD4_SFC_DISTR_VERBOSE
        write(1000+rank,*) 'C allgatherv: rank ',crank, grank
        write(1000+rank,*) 'C allgatherv: cgborders_outside_group ',cgborders_outside_group
#endif
        if( (crank>=cgborders(3) .and. crank<=cgborders(4)) .or. cgborders_outside_group ) then
          ! this rank owns block weights which are part of new coarse partition
          ! first group rank owns all block weights before group
          if(grank==0) then
            start_pos = cgborders(1)
          else
            start_pos = MAX(partition(crank), cgborders(1))
          end if
#ifdef FD4_SFC_DISTR_VERBOSE
          write(1000+rank,*) 'C allgatherv: first_rank_in_cgpart ',start_pos == cgborders(1)
#endif
          ! first rank in a coarse partition additionally sends previous block
          if( start_pos == cgborders(1) ) start_pos = start_pos - 1
          ! last group rank owns all block weights after group
          if(grank==groupsz-1) then
            num_elem = cgborders(2) - start_pos + 1
          else
            num_elem = MIN(partition(crank+1)-1, cgborders(2)) - start_pos + 1
          end if
#ifdef FD4_SFC_DISTR_VERBOSE
          if(num_elem<=0) write(1000+rank,*) 'C allgatherv: num_elem<=0 ',num_elem
#endif
          num_elem = MAX(0, num_elem)
          mpi%recvcnt(grank) = num_elem
#ifdef FD4_SFC_DISTR_VERBOSE
          write(1000+rank,*) 'C allgatherv: l pos ',start_pos-cgborders(1)+1, start_pos-cgborders(1)+1+num_elem-1
          write(1000+rank,*) 'C allgatherv: g pos ',start_pos, start_pos+num_elem-1
#endif
          if(crank == rank .and. .not. cgborders_outside_group) then
            ! local copy
            ! only consider weights of own blocks, not those that were received with p2p msg from other group ranks
            start_pos = MAX(partition(crank), cgborders(1))
            if( start_pos == cgborders(1) ) start_pos = start_pos - 1
            num_elem = MIN(partition(crank+1)-1, cgborders(2)) - start_pos + 1
#ifdef FD4_SFC_DISTR_VERBOSE
            if(num_elem<=0) write(1000+rank,*) 'C allgatherv: local copy num_elem<=0 ',num_elem
#endif
            num_elem = MAX(0, num_elem)
#ifdef FD4_SFC_DISTR_VERBOSE
            write(1000+rank,*) 'C allgatherv: local copy ',num_elem, local_copy_num, local_copy_start
#ifdef FD4_SFC_DISTR_VERBOSE_SYNC
            flush(1000+rank)
            ferr=fsync(fnum(1000+rank))
#endif
#endif
            !FD4_L2_INFO(i2s(num_elem)//i2s(local_copy_num))
            FD4_L1_ASSERT(num_elem/=local_copy_num,'num_elem/=local_copy_num')
            if(num_elem>0) then
              blkweights_ps(start_pos-cgborders(1)+1:start_pos-cgborders(1)+num_elem) = &
                bal%sfcd_blkweights_ps(local_copy_start:local_copy_start+num_elem-1)
            end if
          end if
        else
          mpi%recvcnt(grank) = 0
        end if
        if(grank>0) mpi%recvdis(grank) = mpi%recvdis(grank-1) + mpi%recvcnt(grank-1)
#ifdef FD4_SFC_DISTR_VERBOSE
        write(1000+rank,*) 'C allgatherv: recvcnt ',mpi%recvcnt(grank)
        write(1000+rank,*) 'C allgatherv: recvdis ',mpi%recvdis(grank)
#endif
        crank = crank + 1
      end do

#ifdef FD4_SFC_DISTR_VERBOSE_SYNC
      flush(1000+rank)
      ferr=fsync(fnum(1000+rank))
#endif

      if(mpi%recvcnt(groupsz-1)+mpi%recvdis(groupsz-1)/=ncgblocks+1) then
        FD4_L1_ERROR('recvcnt / recvdis error '//i2s(ncgblocks))
      end if
      
#ifdef FD4_SFC_DISTR_DEBUG
      !!DEBUG: check if all ranks in a group have same recvcnt for MPI_Allgatherv
      call mpi_allreduce(mpi%recvcnt(0),mpi%recvdis(0),groupsz,MPI_I_K,MPI_MAX,bal%sfcd_mpi%comm,err)
      if(ANY(mpi%recvcnt(0:groupsz-1)/=mpi%recvdis(0:groupsz-1))) then
        print *,rank,mpi%recvcnt(0:groupsz-1)
        print *,rank,mpi%recvdis(0:groupsz-1)
        FD4_L1_ERROR('recvcnt error')
      end if
      ! restore recvdis
      mpi%recvdis(0) = 0
      do grank = 1, groupsz-1
        mpi%recvdis(grank) = mpi%recvdis(grank-1) + mpi%recvcnt(grank-1)
      end do
#endif


      ! finally exchange the block weights within the group
      
      if(domain%stat_enabled) call gettime_ofday(timing_us0mpi)

      call MPI_Allgatherv(MPI_IN_PLACE, 0, MPI_DATATYPE_NULL, blkweights_ps(0), &
                          mpi%recvcnt(0),  mpi%recvdis(0), MPI_R8K, bal%sfcd_mpi%comm, err )
      err = err - MPI_SUCCESS
      FD4_L3_ASSERT(err/=0,'MPI_Allgatherv failed')

      if(domain%stat_enabled) then
        call gettime_ofday(timing_us1)
        domain%stat%values(bal%stat_mpi_us) = domain%stat%values(bal%stat_mpi_us) + (timing_us1-timing_us0mpi)
      end if


      ! create local prefix sum (i.e. substract first value from all)
      blkweights_ps(1:ncgblocks) = blkweights_ps(1:ncgblocks) - blkweights_ps(0)
      blkweights_ps(0) = 0.0_r8k
      

#ifdef FD4_SFC_DISTR_COMP_WGT
      !!DEBUG: compare blocks weights with those sent by the other method
      !!DEBUG: (only if blkweights_ps is different array!)
      do ib=0,ncgblocks
        if(blkweights_ps(ib) /= bal%blkweights_ps(ib)) then
          print *,rank,' wgt err at',ib,ib+cgborders(1)-1,blkweights_ps(ib),bal%blkweights_ps(ib)
          err = -42
          return
        end if
      end do
#endif

    end if ! ncgblocks>0
    

    err = 0
       
  end subroutine fd4_part_sfcd_exch_weights_coll


#endif

end module fd4_part_sfcd_mod
